"""
Unit test for processors.py
"""
import pytest

from melusine.processors import (
    DateProcessor,
    DeterministicTextFlagger,
    Message,
    Normalizer,
    RegexTokenizer,
    Segmenter,
    TextExtractor,
    TokensExtractor,
    TransferredEmailProcessor,
)


@pytest.mark.parametrize(
    "input_text, lowercase, output_text",
    [
        ("Héllö WORLD", True, "hello world"),
        ("Hèllo WÖRLD", False, "Hello WORLD"),
        ("", False, ""),
    ],
)
def test_normalizer(input_text, lowercase, output_text):
    """Unit test"""

    normalizer = Normalizer(lowercase=lowercase)
    text = normalizer.normalize_text(input_text)
    assert text == output_text

    assert normalizer.normalize_text(1.25) == ""


def test_normalizer_messages():
    """Unit test"""

    message_1 = Message(meta="", text="Héllö WORLD")
    message_2 = Message(meta="abcd", text="Héllö heLLo")
    normalizer = Normalizer(lowercase=True, input_columns="messages")
    message_list = normalizer.normalize_message([message_1, message_2])

    assert message_list[0].clean_text == "hello world"
    assert message_list[1].clean_text == "hello hello"


@pytest.mark.parametrize(
    "input_text, output_tokens, lowercase, normalization_form",
    [
        ("le petit Chat", ["petit", "chat"], True, None),
        ("le Géant", ["géant"], True, None),
        ("le Géant", ["Geant"], False, "NFKD"),
        ("Comme un grand", ["Comme", "grand"], False, None),
        ("le un et je", [], False, None),
    ],
)
def test_tokenizer(input_text, output_tokens, lowercase, normalization_form):
    """Unit test"""

    tokenizer = RegexTokenizer(
        tokenizer_regex=r"\w+(?:[\?\-\"_]\w+)*",
        stopwords=["le", "un", "et", "je"],
        lowercase=lowercase,
        normalization_form=normalization_form,
    )

    tokens = tokenizer.tokenize(input_text)
    assert tokens == output_tokens


@pytest.mark.parametrize(
    "input_text, expected_messages",
    [
        ("Hello World", [Message(meta="", text="Hello World")]),
        (
            "Merci\nDe : jean@gmail.com\nObjet: Votre attestation\nVoici l'attestation",
            [
                Message(meta="", text="Merci"),
                Message(
                    meta="De : jean@gmail.com\nObjet: Votre attestation",
                    text="Voici l'attestation",
                ),
            ],
        ),
        (
            "Merci\nObjet: Votre attestation\nVoici l'attestation",
            [
                Message(meta="", text="Merci\nObjet: Votre attestation\nVoici l'attestation"),
            ],
        ),
        (
            "Merci\nDe : jean@gmail.com\nSujet : ABCD\nObjet: Votre attestation\nVoici l'attestation",
            [
                Message(meta="", text="Merci"),
                Message(
                    meta="De : jean@gmail.com\nSujet : ABCD\nObjet: Votre attestation",
                    text="Voici l'attestation",
                ),
            ],
        ),
        (
            "Je vous ai Envoyé :\n- le devis\nla facture\nSalutations",
            [
                Message(meta="", text="Je vous ai Envoyé :\n- le devis\nla facture\nSalutations"),
            ],
        ),
        (
            "Message 1\nLe 2 févr. 2022 à 09:10,\ntest@maif.fr\na écrit :\n\n\nMessage 2",
            [
                Message(meta="", text="Message 1"),
                Message(meta="Le 2 févr. 2022 à 09:10,\ntest@maif.fr\na écrit :", text="Message 2"),
            ],
        ),
        (
            "Message 1\nmail transféré\n------------\nLe 2 févr. 2022 à 09:10,\ntest@maif.fr\na écrit :\n\n\nMessage 2",
            [
                Message(meta="", text="Message 1"),
                Message(
                    meta="mail transféré\n------------\nLe 2 févr. 2022 à 09:10,\ntest@maif.fr\na écrit :",
                    text="Message 2",
                ),
            ],
        ),
    ],
)
def test_segmenter(input_text, expected_messages):
    """Unit test"""

    segmenter = Segmenter()
    result = segmenter.segment_text(input_text)
    for i, message in enumerate(result):
        assert message.meta == expected_messages[i].meta
        assert message.text == expected_messages[i].text


@pytest.mark.parametrize(
    "input_message_list, expected_text",
    [
        (
            [
                Message(meta="", text="Hello world"),
                Message(
                    meta="",
                    text="Hello world",
                ),
            ],
            "Hello world",
        ),
        (
            [
                Message(meta="", text="Merci"),
                Message(
                    meta="De : jean@gmail.com\nObjet: Votre attestation",
                    text="Voici l'attestation",
                ),
            ],
            "Merci",
        ),
        (
            [
                Message(meta="", text="Merci", tags=[("THANKS", "Merci")]),
            ],
            "Merci",
        ),
    ],
)
def test_text_extractor(input_message_list, expected_text):
    """Unit test"""

    extractor = TextExtractor(output_columns="text", n_messages=1)
    result = extractor.extract(input_message_list)
    assert result == expected_text


def test_text_extractor_error():
    """Unit test"""
    with pytest.raises(ValueError):
        _ = TextExtractor(output_columns="text", n_messages=1, include_tags=["A"], exclude_tags=["B"])


def test_text_extractor_multiple_messages():
    """Unit test"""
    message_list = [
        Message(meta="", text="", tags=[("BODY", "A"), ("GREETINGS", "G"), ("BODY", "A")]),
        Message(meta="", text="", tags=[("BODY", "B"), ("BODY", "B"), ("BODY", "B")]),
        Message(meta="", text="", tags=[("GREETINGS", "G"), ("BODY", "C"), ("BODY", "C")]),
    ]
    expected_output = "A\nB\nB\nB"

    extractor = TextExtractor(
        output_columns="text",
        n_messages=None,
        stop_at=["GREETINGS"],
        include_tags=["BODY"],
    )
    result = extractor.extract(message_list)
    assert result == expected_output


def test_text_extractor_with_tags():
    """Unit test"""
    input_message_list = [
        Message(meta="", text="Bonjour\nblahblah\nMerci"),
        Message(meta="", text="Bonjour2\nMerci2"),
    ]
    input_message_list[0].tags = [("HELLO", "Bonjour"), ("CUSTOM_TAG", "blahblah"), ("THANKS", "Merci")]
    input_message_list[1].tags = [("HELLO", "Bonjour2"), ("THANKS", "Merci2")]

    extractor = TextExtractor(
        output_columns="text",
        exclude_tags=["HELLO", "CUSTOM_TAG"],
    )
    result = extractor.extract(input_message_list)
    assert result == "Merci"


def test_token_extractor():
    """Unit test"""
    separator = "PAD"
    pad_size = 2

    token_extractor = TokensExtractor(
        input_columns=("body_tok", "header_tok"),
        output_columns=("toks,"),
        pad_size=pad_size,
        sep_token=separator,
    )

    data = {
        "body_tok": ["my", "body"],
        "header_tok": ["my", "header"],
    }

    extracted = token_extractor.extract(data)

    assert extracted == ["my", "body", separator, separator, "my", "header"]


@pytest.mark.parametrize(
    "input_text, output_text",
    [
        ("appelle moi au 0606060606", "appelle moi au flag_phone"),
        ("Tel:0606060606", "Tel: flag_phone"),
        ("ecris moi a l'adresse test@domain.com", "ecris moi a l'adresse flag_email"),
        ("nada nothing rien", "nada nothing rien"),
        ("", ""),
    ],
)
def test_text_flagger_default(input_text, output_text):
    """Unit test"""
    text_flags = {
        "numeric_flags": {r"\d{10}": "flag_phone"},
        r"\w+@\w+\.\w{2,4}": "flag_email",
    }
    text_flagger = DeterministicTextFlagger(text_flags=text_flags)
    text = text_flagger.flag_text(input_text)

    assert text == output_text


@pytest.mark.parametrize(
    "input_text, output_text, add_spaces, remove_multiple_spaces",
    [
        ("Tel:0606060606", "Tel:flag_phone", False, True),
        ("Tel: 0606060606", "Tel: flag_phone", True, True),
        ("Tel: 0606060606", "Tel:  flag_phone", True, False),
        ("", "", False, False),
    ],
)
def test_text_flagger_args(input_text, output_text, add_spaces, remove_multiple_spaces):
    """Unit test"""
    text_flags = {
        "numeric_flags": {r"\d{10}": "flag_phone"},
    }
    text_flagger = DeterministicTextFlagger(
        text_flags=text_flags,
        add_spaces=add_spaces,
        remove_multiple_spaces=remove_multiple_spaces,
    )
    text = text_flagger.flag_text(input_text)

    assert text == output_text


def test_date_processor_instanciation():
    """Test instantiate processor"""
    _ = DateProcessor(input_columns="date", output_columns="date")


@pytest.mark.parametrize(
    "date_str, expected_iso_format",
    [
        ("jeudi 29 avril 2021 21:07", "2021-04-29"),
        ("jeudi 9 avril 2021 21:07", "2021-04-09"),
        ("mar. 11 mai 2021 à 14:29", "2021-05-11"),
        ("4 mai 2021 à 18:14:26 UTC+2", "2021-05-04"),
        ("Lundi 04 janvier 2021 11:35", "2021-01-04"),
        ("mié, 16 dic 2020 a las 16:00", "2020-12-16"),
        ("mié, 16 dic, 2020 a las 16:00", "2020-12-16"),
        ("gio 9 set 2021 alle ore 16:16", "2021-09-09"),
        ("woensdag 16 december 2020 10:01", "2020-12-16"),
        ("lundi 19 juihet 2021, 15:25:42 utc+2", "2021-07-19"),
        ("16 june 2021 14:38", "2021-06-16"),
        ("5 july 2021 at 16:49:41 cest", "2021-07-05"),
        ("sunday, 13 jun 2021", "2021-06-13"),
        ("sunday, 13 jun, 2021", "2021-06-13"),
        ("26 eki 2020 pzt 14:27", "2020-10-26"),
        ("vendredi 6 février 2019 09 h 02", "2019-02-06"),
    ],
)
def test_date_processor(date_str: str, expected_iso_format: str) -> None:
    """Simple base tests"""
    date_processor = DateProcessor()
    date_iso_format: str = date_processor.parse_date_to_iso(date_str)

    assert date_iso_format == expected_iso_format


@pytest.mark.parametrize(
    "message_list, tags_to_ignore, expected_len_message_list, expected_original_from",
    [
        # Direct transfer
        pytest.param(
            [
                Message(
                    meta="De: test.test@test.fr <test.test@test.fr>\n"
                    "Envoyé: mercredi 4 mai 2022 11:11\n"
                    "A: avocats@test.fr; BOB Morane <bob.morane@test.fr>\n"
                    "Objet: dossier Test ,\n",
                    text="bla bla bla",
                    tags=[("BODY", "bla bla bla")],
                )
            ],
            "('FOOTER', 'SIGNATURE')",
            1,
            "test.test@test.fr",
            id="direct_transfer",
        ),
        # Direct transfer with Footer
        pytest.param(
            [
                Message(
                    meta="",
                    text="Envoyé depuis mon Iphone",
                    tags=[("FOOTER", "Envoyé depuis mon Iphone")],
                ),
                Message(
                    meta="De: test.test@test.fr <test.test@test.fr>\n"
                    "Envoyé: mercredi 4 mai 2022 11:11\n"
                    "A: avocats@test.fr; BOB Morane <bob.morane@test.fr>\n"
                    "Objet: dossier Test ,\n",
                    text="bla bla bla",
                    tags=[("BODY", "bla bla bla")],
                ),
            ],
            "('FOOTER', 'SIGNATURE')",
            1,
            "test.test@test.fr",
            id="direct_transfer_with_footer",
        ),
        # Direct transfer with Signature (pattern De:\ntest@gmail\nA:\n)
        pytest.param(
            [
                Message(
                    meta="",
                    text="Jane Doe\n4 rue des oliviers 75001 Ville",
                    tags=[
                        ("SIGNATURE", "4 rue des oliviers 75001 Ville"),
                    ],
                ),
                Message(
                    meta="De :\ntest.test42@test.fr\nEnvoyé :\nvendredi 03 mars 2023 14:28\nÀ :"
                    "\nana@test.fr\nObjet :\nTEST",
                    text="bla bla bla",
                    tags=[("BODY", "bla bla bla")],
                ),
            ],
            "('FOOTER', 'SIGNATURE')",
            1,
            "test.test42@test.fr",
            id="direct_transfer_with_signature",
        ),
        # Direct transfer with multiple Signature patterns
        pytest.param(
            [
                Message(
                    meta="",
                    text="Jane Doe\n4 rue des oliviers 75001 Ville",
                    tags=[
                        ("SIGNATURE_NAME", "Jane Doe"),
                        ("SIGNATURE", "4 rue des oliviers 75001 Ville"),
                    ],
                ),
                Message(
                    meta="De :\ntest.test42@test.fr\nEnvoyé :\nvendredi 03 mars 2023 14:28\nÀ :"
                    "\nana@test.fr\nObjet :\nTEST",
                    text="bla bla bla",
                    tags=[("BODY", "bla bla bla")],
                ),
            ],
            "('FOOTER', 'SIGNATURE')",
            1,
            "test.test42@test.fr",
            id="direct_transfer_with_multiple_signatures",
        ),
        # Other transition pattern (pattern De:\nTest <test@gmail>\nA:\n)
        pytest.param(
            [
                Message(
                    meta=(
                        "De:\nANNA <42test.test@test.fr>\nDate:\n3 mars 2023 à 16:42:50 UTC+1\n"
                        "À:\nbob@test.fr\nObjet:\nTR : 1223456"
                    ),
                    text="bla bla bla",
                    tags=[("BODY", "bla bla bla")],
                ),
            ],
            "('FOOTER', 'SIGNATURE')",
            1,
            "42test.test@test.fr",
            id="transition_pattern_de_date_a",
        ),
        # Other transition pattern (pattern De: "test@gmail"\nA:)
        pytest.param(
            [
                Message(
                    meta=(
                        """De: "test_test@test.fr"\nDate:\n3 mars 2023 à 16:42:50 UTC+1\n"""
                        "À:\nbob@test.fr\nObjet:\nTR : 1223456"
                    ),
                    text="bla bla bla",
                    tags=[("BODY", "bla bla bla")],
                ),
            ],
            "('FOOTER', 'SIGNATURE')",
            1,
            "test_test@test.fr",
            id="transition_pattern_de_a",
        ),
        # Other transition pattern (pattern Le 1 mars..., Abc <abc@gmail.com> a écrit)
        pytest.param(
            [
                Message(
                    meta=("Le 2 mars 2023 à 18:18, Bob <test.test.test@test.fr> a écrit :"),
                    text="bla bla bla",
                    tags=[("BODY", "bla bla bla")],
                ),
            ],
            "('FOOTER', 'SIGNATURE')",
            1,
            "test.test.test@test.fr",
            id="transition_pattern_le_date_text_a_ecrit",
        ),
        # Other transition pattern (pattern Le 01/01/2001 12:12, abc@gmail.com a écrit)
        pytest.param(
            [
                Message(
                    meta=("Le 01/01/2001 11:14, test.test.test@test.fr a écrit :"),
                    text="bla bla bla",
                    tags=[("BODY", "bla bla bla")],
                ),
            ],
            "('FOOTER', 'SIGNATURE')",
            1,
            "test.test.test@test.fr",
            id="transition_pattern_le_date_numbers_a_ecrit",
        ),
        # Other transition pattern (pattern Le dim., 12:12, abc@gmail.com a écrit)
        pytest.param(
            [
                Message(
                    meta=("Le 01/01/2001 11:14, test.test.test@test.fr a écrit :"),
                    text="bla bla bla",
                    tags=[("BODY", "bla bla bla")],
                ),
            ],
            "('FOOTER', 'SIGNATURE')",
            1,
            "test.test.test@test.fr",
            id="transition_pattern_le_weekday_a_ecrit",
        ),
        # Test argument tags_to_ignore (removed Signature from the list)
        pytest.param(
            [
                Message(
                    meta="",
                    text="Jane Doe\n4 rue des oliviers 75001 Ville",
                    tags=[("SIGNATURE", "Jane Doe\n4 rue des oliviers 75001 Ville")],
                ),
                Message(
                    meta="De: test.test@test.fr <test.test@test.fr>\n"
                    "Envoyé: mercredi 4 mai 2022 11:11\n"
                    "A: avocats@test.fr; BOB Morane <bob.morane@test.fr>\n"
                    "Objet: dossier Test ,\n",
                    text="bla bla bla",
                    tags=[("BODY", "bla bla bla")],
                ),
            ],
            "('FOOTER',)",
            2,
            None,
            id="tags_to_ignore",
        ),
        # Not a transfer
        pytest.param(
            [
                Message(
                    meta="",
                    text="J'entends le loup, le renard et la belette",
                    tags=[("BODY", "J'entends le loup, le renard et la belette")],
                ),
                Message(
                    meta="De: test.test@test.fr <test.test@test.fr>\n"
                    "Envoyé: mercredi 4 mai 2022 11:11\n"
                    "A: avocats@test.fr; BOB Morane <bob.morane@test.fr>\n"
                    "Objet: dossier Test ,\n",
                    text="bla bla bla",
                    tags=[("BODY", "bla bla bla")],
                ),
            ],
            "('FOOTER', 'SIGNATURE')",
            2,
            None,
            id="not_a_transfer",
        ),
        # Empty mail
        pytest.param(
            [
                Message(
                    meta="",
                    text="",
                    tags=[("BODY", "")],
                ),
            ],
            "('FOOTER', 'SIGNATURE')",
            1,
            None,
            id="empty_email",
        ),
        # No email address in the meta
        pytest.param(
            [
                Message(
                    meta="",
                    text="Envoyé de mon iPhone",
                    tags=[("FOOTER", "bla 1")],
                ),
                Message(
                    meta="Nothing useful",
                    text="bla 2",
                    tags=[("BODY", "bla 2")],
                ),
            ],
            "('FOOTER', 'SIGNATURE')",
            1,
            None,
            id="missing_email_address",
        ),
    ],
)
def test_transferred_email_processor(message_list, tags_to_ignore, expected_len_message_list, expected_original_from):
    """Unit test"""
    processor = TransferredEmailProcessor(tags_to_ignore=tags_to_ignore, messages_column="message_list")
    processed_message_list, clean_from = processor.process_transfered_mail(message_list=message_list)

    assert clean_from == expected_original_from
    assert len(processed_message_list) == expected_len_message_list



from melusine.regex import (
    EmergencyRegex,
    ReplyRegex,
    ThanksRegex,
    TransferRegex,
    VacationReplyRegex,
)


def test_reply_regex():
    regex = ReplyRegex()
    regex.test()


def test_thanks_regex():
    regex = ThanksRegex()
    regex.test()


def test_transfer_regex():
    regex = TransferRegex()
    regex.test()


def test_vacation_reply_regex():
    regex = VacationReplyRegex()
    regex.test()


def test_emergency_regex():
    regex = EmergencyRegex()
    regex.test()



import logging
import re
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from exchangelib import (  # noqa
    Account,
    Configuration,
    Credentials,
    FaultTolerance,
    FileAttachment,
    Folder,
    HTMLBody,
    Message,
)
from exchangelib.errors import ErrorFolderNotFound  # noqa

logger = logging.getLogger(__name__)


class ExchangeConnector:
    """
    Connector to Outlook Exchange Mailboxs.
    This class contains methods suited for automated emails routing.
    """

    def __init__(
        self,
        mailbox_address: str,
        credentials: Credentials,
        config: Configuration,
        routing_folder_path: Optional[str] = None,
        correction_folder_path: Optional[str] = None,
        done_folder_path: Optional[str] = None,
        target_column: str = "target",
        account_args: Optional[Dict[str, Any]] = None,
        sender_address: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        mailbox_address: str
            Email address of the mailbox. By default, the login address is used
        credentials: Credentials
            Exchangelib credentials to connect to an Exchange mailbox
        config: Configuration
            Exchangelib configuration object
        routing_folder_path: str
            Path of the base routing folder
        correction_folder_path: str
            Path of the base correction folder
        done_folder_path: str
            Path of the Done folder
        target_column: str
            Name of the DataFrame column containing target folder names
        account_args: dict
            Dict containing arguments to instantiate an exchangelib "Account" object.
        sender_address: str
            Email address used to send emails.
        """

        self.sender_address = sender_address
        self.mailbox_address = mailbox_address
        self.folder_list = None
        self.target_column = target_column
        # Default Account parameters
        if not account_args:
            account_args = {"autodiscover": True}

        # Connect to mailbox
        self.credentials = credentials
        self.exchangelib_config = config
        # Mailbox account (Routing, Corrections, etc)
        self.mailbox_account = Account(
            self.mailbox_address,
            credentials=self.credentials,
            config=self.exchangelib_config,
            **account_args,
        )
        # Sender accounts (send emails)
        if sender_address:
            self.sender_account = Account(
                self.sender_address,
                credentials=self.credentials,
                config=self.exchangelib_config,
                **account_args,
            )
            logger.info(f"Address {self.sender_address} is set up to send emails.")
        else:
            self.sender_account = None
            logger.info("Sender address not specified, email sending is disabled.")

        # Setup correction folder and done folder
        self.routing_folder_path = routing_folder_path
        self.correction_folder_path = correction_folder_path
        self.done_folder_path = done_folder_path

        logger.info(f"Connected to mailbox {self.mailbox_address}.")

    def _get_mailbox_path(self, path: Optional[str]) -> Folder:
        """
        Utils function to get a mailbox Folder from a path string.
        Ex:
        - input string : ROUTING
        - output Folder : Folder at root/Haut de la banque d'informations/Boîte de réception/ROUTING

        Parameters
        ----------
        path : str
            String describing the desired path to a mailbox folder

        Returns
        -------
        mailbox_path: Folder
            Mailbox Folder corresponding to the input path
        """
        # Default to inbox
        if not path:
            return self.mailbox_account.inbox

        # Start mailbox path from root folder
        if re.match("/?root/", path, flags=re.I):
            path = re.split("/?root/", path, flags=re.I)[1]
            mailbox_path = self.mailbox_account.root

        # Start mailbox path from inbox folder
        else:
            mailbox_path = self.mailbox_account.inbox

        # Build mailbox path
        folders = path.split("/")
        for folder in folders:
            if folder == "..":
                mailbox_path = mailbox_path.parent
            else:
                mailbox_path = mailbox_path / folder

        return mailbox_path

    @staticmethod
    def _get_folder_path(folder: Folder) -> Union[str, None]:
        """
        Utils function to get the full mailbox path of a folder.
        - input Folder : Folder("Routing")
        - output string : root/Haut de la banque d'informations/Boîte de réception/ROUTING

        Parameters
        ----------
        folder : Folder
            Mailbox folder

        Returns
        -------
        path: str
            Full mailbox path of the input Folder
        """
        if not isinstance(folder, Folder):
            return None

        path = folder.name
        while folder.name != "root":
            folder = folder.parent
            path = folder.name + "/" + path

        return path

    @property
    def routing_folder_path(self) -> Union[str, None]:
        """
        Get the path to the Routing folder.

        Returns
        -------
        path: str
            Path to the Routing folder
        """
        path = self._get_folder_path(self.routing_folder)
        return path

    @routing_folder_path.setter
    def routing_folder_path(self, routing_folder_path: str) -> None:
        """
        Setter for the routing folder.
        """
        self.routing_folder = self._get_mailbox_path(routing_folder_path)
        folder_path = self._get_folder_path(self.routing_folder)
        logger.info(f"Routing folder path set to '{folder_path}'")

    @property
    def done_folder_path(self) -> Union[str, None]:
        """
        Get the path to the Done folder.

        Returns
        -------
        path: str
            Path to the Done folder
        """
        path = self._get_folder_path(self.done_folder)
        return path

    @done_folder_path.setter
    def done_folder_path(self, done_folder_path: str) -> None:
        """
        Setter for the done folder.
        """
        if not done_folder_path:
            self.done_folder = None
            logger.info("Done folder path not set")
        else:
            self.done_folder = self._get_mailbox_path(done_folder_path)
            folder_path = self._get_folder_path(self.done_folder)
            logger.info(f"Done folder path set to '{folder_path}'")

    @property
    def correction_folder_path(self) -> Union[str, None]:
        """
        Get the path to the Correction folder.

        Returns
        -------
        path: str
            Path to the Correction folder
        """
        path = self._get_folder_path(self.correction_folder)
        return path

    @correction_folder_path.setter
    def correction_folder_path(self, correction_folder_path: str) -> None:
        """
        Setter for the correction folder.
        """
        if not correction_folder_path:
            self.correction_folder = None
            logger.info("Correction folder path not set")
        else:
            self.correction_folder = self._get_mailbox_path(correction_folder_path)
            folder_path = self._get_folder_path(self.correction_folder)
            logger.info(f"Correction folder path set to '{folder_path}'")

    def create_folders(self, folder_list: List[str], base_folder_path: Optional[str] = None) -> None:
        """Create folders in the mailbox.

        Parameters
        ----------
        folder_list : list
            Create folders in the mailbox
        base_folder_path : str
            New folders will be created inside at path base_folder_path (Defaults to inbox)
        """
        self.folder_list = folder_list

        # Setup base folder
        base_folder = self._get_mailbox_path(base_folder_path)

        # Check existing folders
        existing_folders = [f.name for f in base_folder.children]

        # Create new folders
        base_folder_name = base_folder_path or "Inbox"
        for folder_name in folder_list:
            if folder_name not in existing_folders:
                f = Folder(parent=base_folder, name=folder_name)
                f.save()
                logger.info(f"Created subfolder {folder_name} in folder {base_folder_name}")

    def get_emails(
        self,
        max_emails: int = 100,
        base_folder_path: Optional[str] = None,
        ascending: bool = True,
    ) -> pd.DataFrame:
        """
        Load emails in the inbox.

        Parameters
        ----------
        max_emails: int
             Maximum number of emails to load
        base_folder_path: str
            Path to folder to fetch
        ascending: bool
            Whether emails should be returned in ascending reception date order

        Returns
        -------
        df_new_emails: pandas.DataFrame
            DataFrame containing nex emails
        """
        logger.info(f"Reading new emails for mailbox '{self.mailbox_address}'")
        base_folder = self._get_mailbox_path(base_folder_path)
        if ascending:
            order = "datetime_received"
        else:
            order = "-datetime_received"

        all_new_data = (
            base_folder.all()
            .only(
                "message_id",
                "datetime_sent",
                "sender",
                "to_recipients",
                "subject",
                "text_body",
                "attachments",
            )
            .order_by(order)[:max_emails]
        )

        new_emails = [self._extract_email_attributes(x) for x in all_new_data if isinstance(x, Message)]
        df_new_emails = pd.DataFrame(new_emails)

        logger.info(f"Read '{len(new_emails)}' new emails")
        return df_new_emails

    @staticmethod
    def _extract_email_attributes(email_item: Message) -> dict:
        """
        Load email attributes of interest such as:
        - `message_id` field
        - `body` field
        - `header` field
        - `date` field
        - `from` field
        - `to` field
        - `attachment` field

        Parameters
        ----------
        email_item: exchangelib.Message
            Exchange Message object

        Returns
        -------
        email_dict: Dict with email attributes of interest
        """
        if not email_item.to_recipients:
            to_list = list()
        else:
            to_list = [i.email_address for i in email_item.to_recipients]

        if not email_item.attachments:
            attachments_list = None
        else:
            attachments_list = [i.name for i in email_item.attachments]

        # Modification to deal with draft in file tree
        if email_item.datetime_sent is not None and email_item.sender is not None:
            email_dict = {
                "message_id": email_item.message_id,
                "body": email_item.text_body or "",
                "header": email_item.subject or "",
                "date": email_item.datetime_sent.isoformat(),
                "from": email_item.sender.email_address or None,
                "to": to_list,
                "attachment": attachments_list,
            }
        else:
            # There is a draft in the emails
            email_dict = {
                "message_id": email_item.message_id,
                "body": email_item.body or "",
                "header": email_item.subject or "",
                "date": None,
                "from": None,
                "to": None,
                "attachment": attachments_list,
            }
        return email_dict

    def route_emails(
        self,
        classified_emails: pd.DataFrame,
        raise_missing_folder_error: bool = False,
        id_column: str = "message_id",
    ) -> None:
        """
        Function to route emails to mailbox folders.

        Parameters
        ----------
        classified_emails: pandas.DataFrame
            DataFrame containing emails message_id and target folder
        raise_missing_folder_error: bool
            Whether an error should be raised when a target folder is missing
        id_column: str
            Name of the DataFrame column containing message ids
        """
        target_column = self.target_column
        target_folders = classified_emails[target_column].unique().tolist()
        base_folder = self.routing_folder

        for folder in target_folders:
            try:
                destination_folder = base_folder / folder
            except ErrorFolderNotFound:
                if raise_missing_folder_error:
                    logger.exception(f"Mailbox (sub)folder '{folder}' not found")
                    raise
                else:
                    logger.warning(f"Mailbox (sub)folder '{folder}' not found")
                    continue

            mask = classified_emails[target_column] == folder
            mids_to_move = classified_emails[mask][id_column]
            items = self.mailbox_account.inbox.filter(message_id__in=mids_to_move).only("id", "changekey")
            self.mailbox_account.bulk_move(ids=items, to_folder=destination_folder, chunk_size=5)
            logger.info(f"Moving {mids_to_move.size} emails to folder '{folder}'")

    def get_corrections(
        self,
        max_emails: int = 100,
        ignore_list: Optional[List[str]] = None,
        correction_column_name: str = "correction",
    ) -> pd.DataFrame:
        """
        When mailbox users find misclassified emails, they should move them to correction folders.
        This method collects the emails placed in the correction folders.

        Parameters
        ----------
        max_emails: Maximum number of emails to fetch at once
        ignore_list: List of folders that should be ignored when fetching emails
        correction_column_name: Name of the column containing the correction folder in the returned DataFrame

        Returns
        -------
        df_corrected_emails: pandas.DataFrame
            DataFrame containing the misclassified emails ids and associated correction folder
        """
        if ignore_list is None:
            ignore_list = []

        if self.correction_folder is None:
            raise AttributeError(
                "You need to set the class attribute `correction_folder_path` to use `get_corrections`."
            )

        logger.info(f"Reading corrected emails from folder and {self.correction_folder}")

        # Get correction folders
        categories = [e.name for e in self.correction_folder.children if e.name not in ignore_list]

        # Load corrected emails
        all_corrected_emails = list()
        for folder_name in categories:
            folder = self.correction_folder / folder_name
            messages = (
                folder.all()
                .only(
                    "message_id",
                    "datetime_sent",
                    "sender",
                    "to_recipients",
                    "subject",
                    "text_body",
                    "attachments",
                )
                .order_by("datetime_received")[:max_emails]
            )
            emails = [self._extract_email_attributes(m) for m in messages if isinstance(m, Message)]

            # Add correction folder to email attributes
            for item in emails:
                item.update({correction_column_name: folder_name})

            all_corrected_emails.extend(emails)
            logger.info(f"Found {len(emails)} corrected emails in folder {folder}")

        logger.info(f"Found {len(all_corrected_emails)} corrected emails in total")
        df_corrected_emails = pd.DataFrame(all_corrected_emails)

        return df_corrected_emails

    def move_to_done(self, emails_id: List[str]) -> None:
        """
        Once the corrected emails have been processed, they can be moved to a "Done" folder.

        Parameters
        ----------
        emails_id: list
            List of emails IDs to be moved to the done folder.
        """
        if (self.correction_folder is None) or (self.done_folder is None):
            raise AttributeError(
                "You need to set the class attribute `done_folder_path` "
                "and the class attribute `correction_folder_path` to use `move_to_done`."
            )
        # Collect corrected emails
        items = self.correction_folder.children.filter(message_id__in=emails_id).only("id", "changekey")
        n_items = items.count()

        # Move to done folder
        self.mailbox_account.bulk_move(ids=items, to_folder=self.done_folder, chunk_size=5)
        logger.info(f"Moved {n_items} corrected emails to the folder {self.done_folder_path}")

    def list_subfolders(self, base_folder_path: Optional[str] = None) -> List[str]:
        """
        List the sub-folders of the specified folder.

        Parameters
        ----------
        base_folder_path: str
            Path to folder to be inspected
        """
        base_folder = self._get_mailbox_path(base_folder_path)
        return [f.name for f in base_folder.children]

    def send_email(self, to: Union[str, List[str]], header: str, body: str, attachments: dict) -> None:
        """
        This method sends an email from the login address (attribute login_address).

        Parameters
        ----------
        to: str or list
            Address or list of addresses of email recipients
        header: str
            Email header
        body: str
             Email body
        attachments: dict
            Dict containing attachment names as key and attachment file contents as values.
            Currently, the code is tested for DataFrame attachments only.
        """
        if self.sender_account is None:
            raise AttributeError(
                "To send emails, you need to specify a `sender_address` when initializing "
                "the ExchangeConnector class."
            )

        if isinstance(to, str):
            to = [to]

        # Prepare Message object
        m = Message(
            account=self.sender_account,
            subject=header,
            body=HTMLBody(body),
            to_recipients=to,
        )
        if attachments:
            for key, value in attachments.items():
                m.attach(FileAttachment(name=key, content=bytes(value, "utf-8")))

        # Send email
        m.send()
        logger.info(f"Email sent from address '{self.sender_address}'")



"""
Define Processors
Processors are objects that can be used as standalone or as steps of a MelusinePipeline.

Implemented classes: [
    Normalizer,
    RegexTokenizer,
    Phraser,
    BaseSegmenter,
    Segmenter,
    BaseExtractor,
    TextExtractor,
    TokensExtractor,
    Tag,
    BaseContentTagger,
    ContentTagger,
    DeterministicTextFlagger,
    Cleaner,
]
"""
from __future__ import annotations

import logging
import re
import unicodedata
from abc import abstractmethod
from re import Pattern
from typing import Any, Iterable, Sequence, Union

import arrow

from melusine.base import MelusineDataset, MelusineTransformer
from melusine.message import Message

logger = logging.getLogger(__name__)


class Normalizer(MelusineTransformer):
    """
    Normalizer transforms raw text into standard text by:
    - Lowercasing
    - Applying unicode normalization standards such as NFD or NFKD
    """

    def __init__(
        self,
        form: str = "NFKD",
        lowercase: bool = True,
        fix_newlines: bool = True,
        input_columns: str = "text",
        output_columns: str = "text",
    ):
        """
        Parameters
        ----------
        form: str
            Unicode normalization form
        lowercase: bool
            If True, lowercase the text
        input_columns: str
            Input columns for the transform operation
        output_columns: str
            Outputs columns for the transform operation
        """

        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
        )

        if "messages" in self.input_columns:
            self.func = self.normalize_message
        else:
            self.func = self.normalize_text

        # Unicode normalization form
        self.form = form

        # Lower casing
        self.lowercase = lowercase

        # Fix newlines
        self.fix_newlines = fix_newlines

    def normalize_message(self, message_list: list[Message]) -> list[Message]:
        """
        Normalize the text of a message.

        Parameters
        ----------
        message_list: list[Message]
            Input message list

        Returns
        -------
        _: list[Message]
            Normalized message list
        """

        for message in message_list:
            message.clean_text = self.normalize_text(message.text)
            message.clean_header = self.normalize_text(message.header)

        return message_list

    def normalize_text(self, text: str) -> str:
        """
        Apply the normalization transformations to the text.

        Parameters
        ----------
        text: str
            Input text to be normalized

        Returns
        -------
        text: str
            Normalized text
        """
        if not isinstance(text, str):
            text = ""

        # Uncommon characters
        text = text.replace("’", "'")
        text = text.replace("œ", "oe")

        # Unicode normalization
        if self.form:
            text = unicodedata.normalize(self.form, text).encode("ASCII", "ignore").decode("utf-8")

        # Lowercasing
        if self.lowercase:
            text = text.lower()

        # Fix newlines
        if self.fix_newlines:
            # Replace platform newlines by standard newline
            text = "\n".join(text.splitlines())

            # Replace multipe spaces by single space
            text = re.sub(r" +", " ", text)

            # Replace multipe newline/spaces patterns by single newline
            text = re.sub(r" *\n+ *", "\n", text)

            # Undesired newlines following quotes
            # Ex : "<\nabc@domain.fr\n>" => "<abc@domain.fr>"
            text = re.sub(r"<\n(\w)", r"<\1", text)
            text = re.sub(r"(\w)\n>", r"\1>", text)

        # Replace non-breaking spaces
        text = text.replace("\xa0", " ")

        return text


class RegexTokenizer(MelusineTransformer):
    """
    Class to split a text into tokens using a regular expression.
    """

    def __init__(
        self,
        tokenizer_regex: str = r"\w+(?:[\?\-\"_]\w+)*",
        stopwords: list[str] | None = None,
        lowercase: bool = True,
        normalization_form: str | None = None,
        input_columns: str = "text",
        output_columns: str = "tokens",
    ):
        """
        Parameters
        ----------
        tokenizer_regex: str
            Regex used to split the text into tokens
        """
        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
            func=self.tokenize,
        )
        # Lowercasing
        self.lowercase = lowercase

        # Normalization
        self.normalization_form = normalization_form

        # Tokenizer regex
        self.tokenizer_regex = tokenizer_regex

        # Stopwords
        if not stopwords:
            self.stopwords = set()
        else:
            self.stopwords = set(stopwords)

    def _text_to_tokens(self, text: str) -> Sequence[str]:
        """
        Method to split a text into a list of tokens.

        Parameters
        ----------
        text: str
            Text to be split

        Returns
        -------
        tokens: Sequence[str]
            List of tokens
        """
        tokens = re.findall(self.tokenizer_regex, text, re.M + re.DOTALL)

        return tokens

    def _remove_stopwords(self, tokens: Sequence[str]) -> Sequence[str]:
        """
        Method to remove stopwords from tokens.

        Parameters
        ----------
        tokens: Sequence[str]
            List of tokens

        Returns
        -------
        tokens: Sequence[str]
            List of tokens without stopwords
        """
        return [token for token in tokens if token not in self.stopwords]

    def tokenize(self, text: str) -> Sequence[str]:
        """
        Method to apply the full tokenization pipeline on a text.

        Parameters
        ----------
        text: str
            Input text to be tokenized

        Returns
        -------
        tokens: Sequence[str]
            List of tokens
        """
        # Lowercase
        if self.lowercase:
            text = text.lower()

        if self.normalization_form:
            text = unicodedata.normalize(self.normalization_form, text).encode("ASCII", "ignore").decode("utf-8")

        # Text splitting
        tokens = self._text_to_tokens(text)

        # Stopwords removal
        tokens = self._remove_stopwords(tokens)

        return tokens


class BaseSegmenter(MelusineTransformer):
    """
    Class to split a conversation into a list of Messages.
    This is an abstract class defining the Segmenter interface.
    Melusine users should implement a subclass or use an existing subclass to segment their emails.
    """

    def __init__(
        self,
        strip_characters: str = "\n >-",
        input_columns: str = "body",
        output_columns: str = "messages",
        regex_flags: re.RegexFlag = re.MULTILINE | re.IGNORECASE,
    ):
        """
        Parameters
        ----------
        strip_characters: str
            Characters to be stripped of text segments
        input_columns: str
            Input columns for the transform operation
        output_columns: str
            Outputs columns for the transform operation
        regex_flags: re.RegexFlag
            Regex flags for segmentation
        """
        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
            func=self.segment_text,
        )

        self.strip_characters = strip_characters

        # Compile segmentation regex
        regex_list = self.create_segmentation_regex_list()
        self._compiled_segmentation_regex = self.compile_regex_from_list(regex_list, flags=regex_flags)

    @staticmethod
    @abstractmethod
    def create_segmentation_regex_list() -> Iterable[str]:
        """
        Method to create a compiled regex that can be used to segment an email.

        Returns
        -------
        _: Iterable[str]
            List of segmentation regexs
        """

    @staticmethod
    def compile_regex_from_list(regex_list: Iterable[str], flags: int | re.RegexFlag = re.M) -> Pattern:
        """
        Method to create a meta-regex from a list of regexs.

        Parameters
        ----------
        regex_list: Iterable[str]
            List of individual regexs
        flags:  int | RegexFlag
            Regex flags

        Returns
        -------
        _:  Pattern[AnyStr]
            Compiled meta-regex
        """
        regex_list = ["(?:" + r + ")" for r in regex_list]
        regex = "|".join(regex_list)

        # Add an overall capture group
        regex = "(" + regex + ")"

        return re.compile(regex, flags=flags)

    def create_messages(self, match_list: list[str]) -> list[Message]:
        """
        Method to create Message instances based on the segmented email data.

        Parameters
        ----------
        match_list: list[str]
            List of text elements matched by the segmentation regex

        Returns
        -------
        _: list[Message]
        """
        # Create first message meta based on email meta
        first_message_meta = ""

        # Strip characters
        match_list = [x.strip(self.strip_characters) for x in match_list]

        # Case email starts with a transition pattern
        if match_list[0] == "":
            if len(match_list) > 1:
                # Adapt first message meta
                first_message_meta = match_list[1]

                # Skip first 2 indices (1st text + 1st meta pattern)
                match_list = match_list[2:]

            else:
                # Empty message
                return [Message(text="")]

        # Insert placeholder for the first message meta
        match_list.insert(0, "")

        # Even indices are meta patterns
        meta_list = [x for i, x in enumerate(match_list) if i % 2 == 0]

        # Odd indices are text
        text_list = [x for i, x in enumerate(match_list) if i % 2 == 1]

        # Replace first message meta
        meta_list[0] = first_message_meta

        return [Message(text=text, meta=meta) for text, meta in zip(text_list, meta_list)]

    def segment_text(self, text: str) -> list[Message]:
        """
        Method to segment a conversation by splitting the text on  transition patterns.
        Ex:
        > Input : "Thank you\nSent by Mr Smith\nHello\nSee attached the document.\nBest Regards"
        > Output :
            - Message(text="Thank you", meta=None)
            - Message(text="Hello\nSee attached the document.\nBest Regards", meta="Sent by Mr Smith")

        Parameters
        ----------
        text: str
            Full text conversation

        Returns
        -------
        _: list[Message]
            List of messages
        """
        # Strip start / end characters
        text = text.strip(self.strip_characters)

        # Split text using the compiled segmentation regex
        matches = self._compiled_segmentation_regex.split(text)

        return self.create_messages(matches)


class Segmenter(BaseSegmenter):
    """
    Class to split a conversation into a list of Messages.
    Inherits from BaseSegmenter.
    Implement methods to segment french emails.
    """

    def __init__(
        self,
        strip_characters: str = "\n >-",
        input_columns: str = "body",
        output_columns: str = "messages",
    ):
        """
        Parameters
        ----------
        strip_characters: str
            Characters to be stripped of text segments
        input_columns: str
            Input columns for the transform operation
        output_columns: str
            Outputs columns for the transform operation
        """
        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
            strip_characters=strip_characters,
        )

    @staticmethod
    def create_segmentation_regex_list() -> Iterable[str]:
        """
        Method to create a compiled regex that can be used to segment an email.

        Returns
        -------
        _: Iterable[str]
            List of segmentation regexs
        """
        # Meta patterns of the form "META_KEYWORD : META_CONTENT"
        # Ex: "De : jean@gmail.com"
        meta_keywords_list_with_semicolon = [
            r"Date",
            r"De",
            r"Exp[ée]diteur",
            r"[ÀA]",
            r"Destinataire",
            r"Envoy[ée](?: le| par)?",
            r"R[ée]pondre [àa]",
            r"Pour",
            r"From",
            r"To",
            r"Sent",
            r"Cc",
        ]
        piped_keywords_with_semicolon = "(?:" + "|".join(meta_keywords_list_with_semicolon) + ")"
        starter_pattern_with_semicolon = rf"^.{{,5}}(?:{piped_keywords_with_semicolon} ?\n? ?: *\n?)"

        # Meta patterns of the form "META_KEYWORD"
        # Ex: "Transféré par jean@gmail.com"
        # (pas de ":")
        # Ex: "------ Message transmis ------"

        regex_weekdays = (
            r"(?:[Ll]undi|[Ll]un\.|[Mm]ardi|[Mm]ar\.|[Mm]ercredi|[Mm]er\.|[Jj]eudi|[Jj]eu\.|"  # noqa
            r"[Vv]endredi|[Vv]en\.|[Ss]amedi|[Ss]am\.|[Dd]imanche|[Dd]im\.)"  # noqa
        )
        regex_months = (
            r"(?:[Jj]anvier|[Ff][ée]vrier|[Mm]ars|[Aa]vril|[Mm]ai|[Jj]uin|[Jj]uillet|"  # noqa
            r"[Aa]o[ûu]t|[Ss]eptembre|[Oo]ctobre|[Nn]ovembre|[Dd][eé]cembre|"  # noqa
            r"(?:janv?|f[ée]vr?|mar|avr|juil?|sept?|oct|nov|d[ée]c)\.)"
        )

        meta_keywords_list_without_semicolon = [
            # Le 2021-01-02 11:20 jane@gmail.fr a écrit :
            # Le 02 juillet 1991 à 11:20 jane@gmail.fr a écrit :
            # Le mardi 31 août 2021 à 11:09, <ville@maif.fr> a écrit :
            (
                rf"\bLe (?:"
                rf"\d{{2}}/\d{{2}}/\d{{4}}|\d{{4}}-\d{{2}}-\d{{2}}|{regex_weekdays}|"  # noqa
                rf"\d{{1,2}} {regex_months})(?:.|\n){{,30}}\d{{2}}:\d{{2}}(?:.|\n){{,50}}(?:\<.{{,30}}\>.{{,5}})?\ba [éecrit]"  # noqa
            ),
            r"Transf[ée]r[ée] par",
            r"D[ée]but du message transf[ée]r[ée] :",
            r"D[ée]but du message r[ée]exp[ée]di[ée] :",
            r"Message transmis",
            r"(?:Message|[Mm]ail) transf[ée]r[ée]",
            r"(?:Courriel|Message|Mail) original",
            r"(?:Message|Mail|Courriel) d'origine",
            r"Original [Mm]essage",
            r"Forwarded message",
            r"Forwarded by",
        ]
        piped_keywords_without_semicolon = "(?:" + "|".join(meta_keywords_list_without_semicolon) + ")"  # noqa
        starter_pattern_without_semicolon = f"{piped_keywords_without_semicolon}(?:[\n ]*--+)?"

        # Combine pattern with and without semicolon
        starter_pattern = rf"(?:{starter_pattern_with_semicolon}|{starter_pattern_without_semicolon})"  # noqa

        # Match everything until the end of the line.
        # Match End of line "\n" and "space" characters
        end_pattern = r".*[\n ]*"

        # Object / Subject pattern (These patterns are not sufficient to trigger segmentation)
        object_line_pattern = "(?:^.{,5}(?:Objet|Subject|Sujet) ?\n? ?: *\n?)" + end_pattern

        # Starters are separated from the meta-values by
        # a semicolon and optional spaces/line breaks
        full_generic_meta_pattern = rf"(?:{starter_pattern}{end_pattern}(?:{object_line_pattern})*)+"

        # Make a tuple of patterns
        pattern_list = (full_generic_meta_pattern,)

        return pattern_list


class BaseExtractor(MelusineTransformer):
    """
    Class to extract data from a list of messages.
    This is an abstract class defining the interface for extractor classes.
    Melusine users should implement a subclass (or use an existing subclass) to extract data from a list of messages.
    """

    def __init__(
        self,
        input_columns: str | Iterable[str],
        output_columns: str,
    ):
        """
        Parameters
        ----------
        input_columns: Union[str, Iterable[str]]
            Input columns for the transform operation
        output_columns: str
            Outputs columns for the transform operation
        """
        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
            func=self.extract,
        )

    @abstractmethod
    def extract(self, message_list: list[Message]) -> Any:
        """
        Method to extract data from a list of messages.

        Parameters
        ----------
        message_list: list[Message]
            List of Messages

        Returns
        -------
        _: str
            Extracted text
        """


class TextExtractor(BaseExtractor):
    """
    Class to extract text data from a list of messages.
    """

    def __init__(
        self,
        input_columns: str = "messages",
        output_columns: str = "last_message",
        include_tags: list[str] | None = None,
        exclude_tags: list[str] | None = None,
        sep: str = "\n",
        n_messages: int | None = 1,
        stop_at: Iterable[str] = ("GREETINGS",),
    ):
        """
        Parameters
        ----------
        input_columns: str
            Input columns for the transform operation
        output_columns: str
            Outputs columns for the transform operation
        include_tags: list[str]
            Message tags to be included in the text extraction
        exclude_tags: list[str]
            Message tags to be excluded from the text extraction
        sep: str
            Separation symbol to join text parts
        n_messages: Union[int, None]
            Number of messages to take into account (starting with the latest)
        stop_at: list[str]
            When stop_at tags are encountered, stop extracting text of the message
        """
        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
        )

        if include_tags and exclude_tags:
            raise ValueError(f"{type(self).__name__} :" "You should specify only of include_tags/exclude_tags")

        self.include_tags = include_tags
        self.exclude_tags = exclude_tags
        self.sep = sep
        self.n_messages = n_messages
        self.stop_at = stop_at

    def extract(self, message_list: list[Message]) -> str:
        """
        Method to extract text parts from a list of messages.

        Parameters
        ----------
        message_list: list[Message]
        Input message list

        Returns
        -------
        _: str
            Extracted text
        """
        if self.n_messages is None:
            n_messages = len(message_list)
        else:
            n_messages = self.n_messages

        text_list = list()

        for message in message_list[:n_messages]:
            # Message has been tagged
            if message.tags is not None:
                if self.include_tags:
                    tags = message.extract_parts(target_tags=self.include_tags, stop_at=self.stop_at)
                    message_text_list = [x[1] for x in tags]
                elif self.exclude_tags:
                    tags = message.extract_parts(target_tags=None, stop_at=self.stop_at)
                    message_text_list = [part for tag, part in tags if tag not in self.exclude_tags]
                else:
                    message_text_list = [part for tag, part in message.tags]

                # Join message text list
                extracted_text = self.sep.join(message_text_list)

            # Message has not been tagged
            else:
                extracted_text = message.text

            text_list.append(extracted_text)

        return self.sep.join(text_list).strip()


class TokensExtractor(BaseExtractor):
    """
    Class to extract tokens from different DataFrame columns.
    Ex:
    > (input column 1) body_tokens: ["hello", "how", "are", "you"]
    > (input column 2) header_tokens: ["catch", "up"]
    > (output column) all_tokens: ["catch", "up", "hello", "how", "are", "you"]

    """

    def __init__(
        self,
        input_columns: str | Iterable[str] = ("header_tokens", "body_tokens"),
        output_columns: str = "tokens",
        sep_token: str = "[PAD]",
        pad_size: int = 5,
    ):
        """
        Parameters
        ----------
        input_columns: Union[str, Iterable[str]]
            Input columns for the transform operation
        output_columns: str
            Outputs columns for the transform operation
        """
        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
        )

        self.sep_token = sep_token
        self.pad_size = pad_size

    def extract(self, row: MelusineDataset) -> list[str]:
        """
        Method to extract tokens from different columns of a DataFrame.

        Parameters
        ----------
        row: MelusineDataset
            Emails input data

        Returns
        -------
        _: list[str]
            List of extracted tokens
        """

        pad_pattern = [self.sep_token] * self.pad_size
        tokens = list()
        for col in self.input_columns:
            tokens += row[col]
            tokens += pad_pattern

        # Remove trailing padding tokens
        tokens = tokens[: -self.pad_size]

        return tokens


class Tag(property):
    """
    Class used by the ContentTagger to identify text tags such as:
    - BODY
    - HELLO
    - GREETINGS
    """


TagPattern = Union[str, Iterable[str], re.Pattern]


class BaseContentTagger(MelusineTransformer):
    """
    Class to add tags to a text
    This is an abstract class defining the interface for all ContentTaggers.
    Melusine users should implement a subclass (or use an existing subclass) to add tags to texts.
    """

    def __init__(
        self,
        input_columns: str = "messages",
        output_columns: str = "messages",
        tag_list: list[str] | None = None,
        default_tag: str = "BODY",
        valid_part_regex: str = r"[a-z0-9?]",
        default_regex_flag: int = re.IGNORECASE,
        post_process: bool = True,
        text_attribute: str = "text",
    ):
        """
        Parameters
        ----------
        input_columns: str
        output_columns: str
        tag_list: list[str]
            (Ordered) List of tags to look for
        default_tag: str
            Tag given to arbitrary text parts
        default_regex_flag: int
            Default flag to compile regex
        text_attribute: str
            Message attribute containing the text data
        """
        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
            func=self.tag_email,
        )

        # If no tag list is specified, use all the tags
        if not tag_list:
            self.tag_list = self.get_tag_list()
        else:
            self.tag_list = tag_list

        # Sep default tag
        self.default_tag = default_tag

        # Set default flag
        self.default_regex_flag = default_regex_flag

        # Set text attribute
        self.text_attribute = text_attribute

        # Activate post-processing
        self.post_process = post_process

        # Pattern to split text into sentences (=parts)
        self.split_pattern = self.compile_split_pattern()

        # Pattern to validate that a text part is valid
        self.valid_part_regex = valid_part_regex

        # Build the regex_dict
        self.regex_dict = {}
        for tag in self.tag_list:
            self.regex_dict[tag] = self.compile_tag_regex(tag)

    def __getitem__(self, key: str) -> re.Pattern:
        """
        Method to access regex corresponding to individual tags easily.
        Ex:
        > t = ContentTagger()
        > t["HELLO"].match("bonjour")

        Parameters
        ----------
        key: str
            Name of a tag

        Returns
        -------
        _: re.Pattern
            Compiled regex
        """
        return self.regex_dict[key]

    @staticmethod
    def compile_split_pattern() -> re.Pattern:
        """
        Method to compile the sentence split regex pattern.

        Ex:
        Bonjour Mr. Dupont. Salutations
        will be splited using the split pattern into
        ["Bonjour Mr. Dupont", "Salutations"]

        Returns
        -------
        _: re.Pattern
            Compiled regex
        """
        # Dot exception patterns
        _madame_pattern = r"(?<!\bM[Mm]e)"
        _monsieur_pattern = r"(?<!\bMr)"
        _mademoiselle_pattern = r"(?<!\bMelle)"
        _mademoiselle_pattern2 = r"(?<!\bMlle)"
        _doctor_pattern = r"(?<!\bDr)"
        _professor_pattern = r"(?<!\bPr)"
        _acronym_pattern = r"(?<!\b[A-Z])"

        # Hello exception patterns
        hello_split_pattern = r"(?:^[Bb](?:onjour|onsoir)\s*,)+"

        # Regular_split_patterns
        newline_pattern = r"\n+"
        dot_pattern = r"(?:[.?!]+(?![a-zA-Z0-9]))+"

        # Build pattern to avoid undesired splitting on the point character
        dot_pattern_with_negativelookbehind = (
            r"(?:"
            + _madame_pattern
            + _monsieur_pattern
            + _mademoiselle_pattern
            + _mademoiselle_pattern2
            + _doctor_pattern
            + _professor_pattern
            + _acronym_pattern
            + dot_pattern
            + r")"
        )

        sentence_split_pattern = (
            r"( *(?:" + f"{dot_pattern_with_negativelookbehind}|{hello_split_pattern}|{newline_pattern}" + r")+) *"
        )
        return re.compile(sentence_split_pattern)

    @classmethod
    def get_tag_list(cls) -> list[str]:
        """
        Method to get the list of available tags.

        Returns
        -------
        _: list[str]
            List of tags
        """
        return [p for p in dir(cls) if isinstance(getattr(cls, p), Tag)]

    def tag_email(self, messages: list[Message]) -> list[Message] | None:
        """
        Method to apply content tagging on an email (= List of Messages)

        Parameters
        ----------
        messages : list[Message]
            List of messages

        Returns
        -------
        messages : list[Message]
            List of messages after content tagging
        """
        if not messages:
            return None

        for message in messages:
            tags = self.tag_text(getattr(message, self.text_attribute))
            message.tags = tags

        return messages

    def compile_tag_regex(self, tag: str) -> re.Pattern:
        """
        Method to validate and compile the regex associated with the input tag.
        Return an error if the regex pattern is malformed.

        Parameters
        ----------
        tag: str
            Tag of interest

        Returns
        -------
        _: re.Pattern
            compiled regex
        """
        # Collect data associated with the input tag
        if hasattr(self, tag):
            regex = getattr(self, tag)
        else:
            raise ValueError(f"Unknown tag {tag}")

        # If a list is provided, pipe it into a string
        if (not isinstance(regex, str)) and isinstance(regex, Iterable):
            regex = "|".join(regex)

        # Compile regex from string
        if isinstance(regex, str):
            try:
                regex = re.compile(regex, flags=self.default_regex_flag)
            except re.error:
                raise ValueError(f"Invalid regex for tag {tag}:\n{regex}")
        elif isinstance(regex, re.Pattern):
            pass
        else:
            raise ValueError(
                f"Tag {tag} does not return any of the supported types : "
                "str "
                "list[str] "
                "re.Pattern "
                f"Got {type(regex)} instead."
            )

        return regex

    def tag_text(self, text: str) -> list[tuple[str, str]]:
        """
        Method to apply content tagging on a text.

        Parameters
        ----------
        text: str
            Input text

        Returns
        -------
        _: list[tuple[str, str]]
            List of tag/text couples (ex: [("HELLO", "bonjour")])
        """
        parts = self.split_text(text)
        tags = list()
        for part in parts:
            tags.append(self.tag_part(part))

        # Post process tags
        if self.post_process:
            tags = self.post_process_tags(tags)

        return tags

    def split_text(self, text: str) -> list[str]:
        """
        Method to split input text into sentences/parts using a regex.

        Parameters
        ----------
        text: str
            Input text

        Returns
        -------
        _: list[str]
            List of parts/sentences
        """
        # Replace multiple spaces by single spaces
        text = re.sub(r" +", " ", text)

        # Split text into sentences
        parts = self.split_pattern.split(text)

        # Cleanup sentence split
        clean_parts = self.clean_up_after_split(parts)

        return [p.strip() for p in clean_parts if self.validate_part(p)]

    def validate_part(self, text: str) -> bool:
        """
        Method to validate a text part.
        By default, check that it contains at least one of:
        - a letter
        - a number
        - an interrogation mark.

        Parameters
        ----------
        text: Text part to be validated

        Returns
        -------
        _: bool
            True if text part is valid
        """
        return bool(re.search(self.valid_part_regex, text, flags=re.I))

    @staticmethod
    def clean_up_after_split(parts: list[str | None]) -> list[str]:
        """
        Clean up sentences after splitting.
        Typically, put punctuation back at the end of sentences.

        Parameters
        ----------
        parts: list[Union[str, None]]

        Returns
        -------
        clean_parts: list[str]
        """
        clean_parts: list[str] = []
        for part in parts:
            if not part:
                continue

            # Part contains punctuation only
            if (len(clean_parts) > 0) and re.search(r"^[ .?!\n]+$", part):
                # Add characters to the previous part
                clean_parts[-1] += part
                continue

            # Regular part
            clean_parts.append(part.strip("\n"))

        return clean_parts

    def tag_part(self, part: str) -> tuple[str, str]:
        """
        Method to apply tagging on a text chunk (sentence/part).

        Parameters
        ----------
        part: str
            Text chunk

        Returns
        -------
        match_tag: str
            Output tag
        part: str
            Original text
        """
        match_tag = self.default_tag

        for tag, regex in self.regex_dict.items():
            match = regex.match(part)
            if match:
                match_tag = tag
                break

        return match_tag, part

    @staticmethod
    def word_block(n_words: int, word_character_only: bool = False) -> str:
        """
        Method to dynamically generate regex patterns to match block of words.

        Parameters
        ----------
        n_words: int
            Number of words to be matched
        word_character_only: bool
            If True, match word characters only

        Returns
        -------
        _: str
            Regex matching desired pattern
        """
        if word_character_only:
            positive = r"\w"
        else:
            # Non-space characters except - and – (considered word separators)
            positive = r"[^\r\s\t\f\v \-–]"

        return rf"(?:[ \-–]*(?:{positive}+(?:[ \-–]+{positive}+){{,{n_words - 1}}})? *)"

    def __call__(self, text: str) -> list[tuple[str, str, str]]:
        """
        Method to find all regex patterns matching the input text.

        Parameters
        ----------
        text: str
            Text to match

        Returns
        -------
        match_list: list[tuple[str, str]]
            List of matching regexes and associated tags
        """
        full_match_list = list()

        # Split parts of the input string
        parts = self.split_text(text)
        for part in parts:
            for tag in self.regex_dict.keys():
                # Get regex or list of regexes
                regex = getattr(self, tag)

                # Find matching regexes
                matching_regex_list = self.find_matching_regex_patterns(part, regex)

                # Format result
                full_match_list.extend([(part, tag, regex) for regex in matching_regex_list])

        return full_match_list

    def find_matching_regex_patterns(self, part: str, regex: TagPattern) -> list[str]:
        """
        Given a regex string, a regex pattern or a list of regexes.
        Find all matching patterns
        """
        matching_regex_list = []
        # Tag defined with a string regex
        if isinstance(regex, str):
            regex_match = re.match(regex, part, flags=self.default_regex_flag)
            if regex_match:
                matching_regex_list.append(regex)

        # Tag defined with a list of string regexes
        elif isinstance(regex, Iterable):
            for r in regex:
                regex_match = re.match(r, part, flags=self.default_regex_flag)
                if regex_match:
                    matching_regex_list.append(r)

        else:
            regex_match = regex.match(part)
            if regex_match:
                matching_regex_list.append(regex.pattern)

        return matching_regex_list

    @abstractmethod
    def post_process_tags(self, tags: list[tuple[str, str]]) -> list[tuple[str, str]]:
        """
        Method to apply tagging rules posterior to the standard regex tagging.

        Parameters
        ----------
        tags: list[tuple[str, str]]
            Original tags

        Returns
        -------
        _: list[tuple[str, str]]
            Post-processed tags
        """


class ContentTagger(BaseContentTagger):
    """
    Class to add tags to a text.
    This class inherits from the base class BaseContentTagger.
    This class implements Content Tagging for French emails.

    Implemented tags:
    - HELLO
    - PJ
    - DISCLAIMER
    - FOOTER
    - GREETINGS
    - SIGNATURE
    """

    ENGLISH_TIMES = ["day", "morning", "afternoon", "evening", "night", "week(-?end)?"]

    def __init__(
        self,
        input_columns: str = "messages",
        output_columns: str = "messages",
        tag_list: list[str] | None = None,
        default_tag: str = "BODY",
        valid_part_regex: str = r"[a-z0-9?]",
        default_regex_flag: int = re.IGNORECASE | re.MULTILINE,
        post_process: bool = True,
        text_attribute: str = "text",
    ):
        """
        Parameters
        ----------
        input_columns: str
        output_columns: str
        tag_list: list[str]
            (Ordered) List of tags to look for
        default_tag: str
            Tag given to arbitrary text parts
        default_regex_flag: int
            Default flag to compile regex
        text_attribute: str
            Message attribute containing the text data
        """
        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
            tag_list=tag_list,
            default_tag=default_tag,
            valid_part_regex=valid_part_regex,
            default_regex_flag=default_regex_flag,
            post_process=post_process,
            text_attribute=text_attribute,
        )

    @Tag
    def GREETINGS(self) -> str | list[str] | re.Pattern:
        """
        Tag associated with email closure sentences.
        Watchout, this tag typically marks the end of a message.
        Ex: "Cordialement"
        """
        english_times_pattern = "|".join(self.ENGLISH_TIMES)
        return [
            r"^.{0,30}cordialement.{0,30}$",
            r"^.{0,5}sinc[èe]rement.{0,30}$",
            r"^.{0,10}cdl?t.{0,16}$",
            r"^.{0,10}bien aimablement.{0,16}$",
            r"^.{0,10}courtoisement.{0,16}$",
            r"^.{0,10}bien [àa] (?:toi|vous).{0,16}$",
            r"^.{0,10}sentiments? (?:d[ée]vou[ée]s?|mutualistes?).{0,16}$",
            r"^.{0,10}(Veuillez.{,3})?(accepte[zr] .{,8}|receve[zr] .{,8})?(meilleure?s?|sinc[eè]res?|cordiale?s?)? ?(salutations?|sentiments?).{0,16}$",
            r"^.{0,45}(?:(?:l')?expression|assurance) de (?:nos|mes) sentiments.{0,16}$",
            r"^.{0,50}(?:salutations?|sentiments?) distingu.{0,30}$",
            r"^.{0,10}Respectueusement.{0,16}$",
            r"^.{0,20}(?:souhait.{,10})?(?:continuation|r[ée]ception).{0,16}$",
            r"^.{0,20}dans l'attente de (?:votre (?:retour|r[ée]ponse)|vous lire).{0,16}$",
            r"^.{0,30}(?:une )bonne r[ée]ception.{0,16}$",
            r"^.{0,3}Bonne r[ée]ception.{0,3}$",
            r"^.{0,3}votre bien d[ée]vou[ée]e?.{0,3}$",
            r"^.{0,3}amicalement votre.{0,3}$",
            r"^.{,3}je vous prie de croire.{,50}(expression|assurance)?.{,50}(consideration|salutations|sentiments).{,30}$",
            # English
            r"^.{0,3}regards.{0,3}$",
            r"^.{0,3}(best|warm|kind|my) *(regards|wishes)?.{0,3}$",
            r"^.{0,3}(yours)? *(truly|sincere?ly|respectfull?y|faithfully).{0,3}$",
            r"^.{0,3}yours.{0,3}$",
            r"^.{0,3}cheers.{0,3}$",
            "^.{0,3}(talk|write|see you|speak to you) soon.{0,3}$",
            "^.{0,3}take care.{0,3}$",
            "^.{0,3}catch you later.{0,3}$",
            rf"^.{{0,3}}have an? (blessed|excellent|good|fantastic|great) ({english_times_pattern}).{{0,3}}$",
            r"i am looking forward to hearing from you.{0,3}$",
            "^.{0,3}looking forward to your reply.{0,3}$",
            "^.{0,3}hoping to hear from you( soon)?.{0,3}$",
        ]

    @Tag
    def HELLO(self) -> str | list[str] | re.Pattern:
        """
        Tag associated with email opening sentences.
        Sentences that can be either opening or closing should be placed here.
        Ex1: "Bonjour"
        Ex2: "Bonne année"
        """
        english_times_pattern = "|".join(self.ENGLISH_TIMES)

        # === Souhaits de bonheur ===
        # J'espère que vous avez passé un bon week-end, etc
        # Bonne semaine
        bon_list = [
            r"ann[ée]e",
            r"(fin de )?journ[ée]e?",
            r"soir[ée]e?",
            r"week[ -]?end",
            r"we",
            r"nuit",
            r"(fin de )?semaine",
            r"f[eê]tes?",
            r"(fin d')?apr[eè]s[ -]?midi",
        ]

        deb_bon_list = [
            r"Bon(?:ne)?",
            r"Bel(?:le)?",
            r"Beau",
            r"joyeux?(?:se)?",
            r"Excel?lent(?:e)?",
        ]

        hello_words_list = [
            r"ch[èe]r[.()es]{,4}",
            r"bonjour",
            r"bonsoir",
            r"madame",
            r"monsieur",
            r"mesdames",
            r"messieurs",
            # English
            rf"good {english_times_pattern}",
            r"hi( there)?",
            r"hello",
            r"greetings",
            r"dear",
            r"dear (m(rs?|s)\.?|miss|madam|mister|sir)( or (m(rs?|s)\.?|miss|madam|mister|sir))?",
            r"sir",
            r"how are you (doing|today)",
            r"(it is|it's)? ?(good|great) to hear from you",
            r"i hope (you are|you're)( doing)? well",
            rf"i hope (you are|you're) having an? ?(great|wonderful|fantastic)? ({english_times_pattern})",
            r"i hope this email finds you well",
            r"to whom it may concern",
        ]
        hello_pattern = "|".join(hello_words_list)

        return [
            r"^.{0,10}((\b" + hello_pattern + r")\b\s*){1,3}(\w+\b\s*){,4}.{,3}(?!.)$",
            rf"^.{{0,16}}(?:{'|'.join(deb_bon_list)}) \b(?:{'|'.join(bon_list)}).{{0,40}}$",
        ]

    @Tag
    def PJ(self) -> str | list[str] | re.Pattern:
        """
        Tag associated with email attachment mentions.
        Ex: "See attached files"
        """
        return [
            r"\(See attached file: .{,60}?\..{1,4}\)",
            r"\(Embedded image(?: moved to file:)? .{,60}?\)",
            r"^.{,4}[\[\(][a-z0-9-_]*\.[a-zA-Z]{3,4}[\]\)]\s*$",
        ]

    @Tag
    def FOOTER(self) -> str | list[str] | re.Pattern:
        """
        Tag associated with email footer sentences.
        Ex: "Envoyé de mon iPhone"
        """
        prefix = r"^.{0,40}"
        suffix = r".*"

        text_list = [
            # Français
            r"[aà] [l]['’ ]attention de.{0,80}$",
            r"Les informations contenues dans ce courrier [ée]lectronique et toutes les pi[eè]ces qui y sont jointes",
            r"Le pr[ée]sent document est couvert par le secret professionnel",
            r"Ce message et toutes les pi[eè]ces jointes sont confidentiels",
            r"Ce message et les fichiers [ée]ventuels",
            r"Ce message est confidentiel",
            r"Ce message contient des informations confidentielles"
            r"Si vous avez recu ce message ([ée]lectronique )?par erreur",
            r"Toute modification, [ée]dition, utilisation",
            r"Tout usage, communication ou reproduction",
            r"Il ne peut etre lu, ni copi[ée], ni communiqu[ée], ni utilis[ée]",
            r"L'[ée]metteur d[ée]cline toute responsabilit[ée]",
            r"Sauf mention contraire, le present message",
            r"Afin de faciliter nos echanges et optimiser le traitement",
            r"Afin de contribuer au respect de l'environnement",
            r"N'imprimez ce message que si cela est indispensable",
            r"Pensez a l'environnement avant d'imprimer ce message",
            r"Droit a la d[ée]connexion",
            r"Ceci est un mail automatique",
            r"Les formats de fichiers acceptés sont : PDF, DOC, DOCX, JPEG, JPG, TIFF, TXT, ODT, XLS, XLSX",
            r"Tout autre format de fichiers ne sera pas transmis au dossier",
            # English
            r"This message and any attachments are confidential",
            r"This e-mail and any files transmitted",
            r"If you have received this (?:message|email) in error",
            r"Any unauthorized modification",
            r"The sender shall not be liable",
        ]

        diclaimer_regex_list = [f"{prefix}{x}{suffix}" for x in text_list]

        miscellaneous_footer_regex = [
            r"(?:courrier electronique|virus|antivirus){2,}",
            r"^.{0,10}Partag[ée] [aà].{0,5} partir de Word pour \b\w+\b$",
            r"^.{0,10}Provenance : Courrier pour Windows",
            r"^.{0,10}garanti sans virus.{0,30}",
            r"^.{0,10}www.avg.com",
            r"^.{0,10}www.avast.com",
            r"^.{0,10}T[ée]l[ée]charg.{0,10}$",
            r"^.{0,2}Obtenir{0,2}$",
            r"(?:Obtenez|T[ée]l[ée]charge[zr])? ?Outlook pour .*",
            r"^.{0,10}La visualisation des fichiers PDF n[ée]cessite.*",
            r"^.{0,10}Si vous recevez ce message par erreur",
            r"^.{0,10}Retrouvez-nous sur www\.maif-\w+\.fr",
            (
                r"^.{0,10}afin de contribuer au respect de l'environnement, merci de n'imprimer ce courriel qu'en c"
                r"as de n[ée]cessit[ée]"
            ),
            (
                r"^.{0,10}(?:Envoy[ée]|Numeris[ée]|Partag[ée]) de(?:puis)?\s*(?:mon)?\s*(?:mobile|smartphone|appareil|"
                r"\biP.|Galaxy|Yahoo|T[ée]l[ée]phone|(?:l'application))"
            ),
            r"^.{0,25}pour Android.{0,5}$",
            r"^.{0,5}Envoy[ée] avec.{0,10}$",
            r"^.{0,5}Envoy[ée] [àa] partir de.{0,35}$",
            r"^.{0,2}Courrier.{0,2}$",
            r"^.{0,2}Pour Windows.{0,2}$",
            r"^.{0,10}Scann[ée] avec.{,30}$",
            r"^.{,3}sans virus.{,3}$",
            # English
            r"^.{0,5}Sent with .*",
            r"^.{0,5}Sent from my .*",
        ]

        return diclaimer_regex_list + miscellaneous_footer_regex

    @Tag
    def THANKS(self) -> str | list[str] | re.Pattern:
        """
        Tag associated with email thanks sentences.
        Ex: "Merci beaucoup"
        """
        thanks_expressions = [
            r"(re)?(merci(e|ant)?(\s(d'|par)\s?avance)?)",
            r"thanks?( you)?",
            r"thx",
        ]
        thanks_pattern = r"\b(" + "|".join(thanks_expressions) + r")\b"

        exception_expressions = [
            r"de",
            r"d",
            r"mais",
            r"cependant",
            r"par contre",
            r"toutefois",
            r"pourtant",
            r"but",
            r"however",
        ]
        exception_pattern = r" *\b(" + "|".join(exception_expressions) + ") *"

        return [
            r"^.{0,20}" + thanks_pattern + r"(?!.{0,5}" + exception_pattern + r").{0,40}(?!.)",
        ]

    @Tag
    def SIGNATURE(self) -> str | list[str] | re.Pattern:
        """
        Tag associated with email signature sentences.
        Ex: "Tel : 0600000000"
        """

        # Jobs lines
        jobs = [
            r"associations?",
            r"(?:\w+ )?analyste?",
            r"conducteur",
            r"[ée]quipe",
            r"soci[ée]t[ée]",
            r"secr[ée]taires?",
            r"secr[ée]tariats?",
            r"directions?",
            r"services?",
            r"assistante?",
            r"gestionnaire",
            r"technicienn?e?",
            r"conseill[eè]re?",
            r"maitres?",
            r"avocats?",
            r"s\.a\.s\.",
            r"squad",
            r"charg[ée]e? d['e]\s*\w+(?:\s*\w+){,2}",
            # English
            r"Lead",
            r"Chief",
            r"VP",
            r"C.O",
            r"(Sales)? Representative",
        ]
        job_regex = r"\b(" + r"|".join(jobs) + r")\b"
        line_with_known_job = rf"(?:^ *.{{,5}}{self.word_block(1)}{job_regex}( +{self.word_block(6)})?(?:\n+|$))"

        # Street address regex
        street_word_list = [
            r"all[ée]e",
            r"avenue",
            r"boulevard",
            r"chemin",
            r"cours",
            r"[ée]splanade",
            r"h?ameau",
            r"impasse",
            r"lotissement",
            r"passage",
            r"place",
            r"square",
            r"quai",
            r"r[ée]sidence",
            r"rue",
            r"sentier",
            # English
            r"st\.?",
            r"street",
            r"ln\.?",
            r"lane",
            r"rd\.?",
            r"road",
            r"hill",
        ]
        street_word_pattern = r"\b(" + "|".join(street_word_list) + r")\b"

        # A number (house number) or range, free words (up to 2), an equivalent of street (rue, allée, etc)
        # and more free words (up to 5), free chars at the end (up to 2)
        street_address_regex = (
            r"^ *\d+(?:-\d+)?(?:bis|ter|b)?,? +(\w+\b *){,2}\b" + street_word_pattern + r"\b *(\w+\b[ -]*){,5}.{,2}$"
        )

        # Email address
        email_address_regex = r"(?:[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,})"

        return [
            # Phone / Fax
            r"(?:^.{,3}(?:T[ée]l(?:[ée]phone)?\.?|mobile|phone|num[ée]ro|ligne).{,20}(?: *(?:\n+|$)))",
            (
                r"^(.{,10}:? ?\(?((?:\+|00)\(?33\)?(?: ?\(0\))?|0)\s*[1-"
                r"9]([\s.-]*\d{2}){4}.{,10}){,3}" + rf"({email_address_regex}.{{,10}})?"
                "( *(\n+|$))"
            ),
            r"^.{,3}(T[ée]l[ée]?(phone|copie)?|Fax|mobile|phone|num[ée]ro|ligne).{,20}$",
            r"^.{,3}Appel non surtax[ée].{,3}$",
            # Street / Address / Post code
            street_address_regex,
            r"^.{,3}Adresse.{,3}$",
            r"(?: *(?:BP|Boite Postale) *\d{,6} .{,30}(?: *(?:\n|$)))",
            r"(?: *\b\d{5}\b(?: *(?:\n|$))?(?: *(?:\S+(?: +\S+){,5})? *)(?: *(?:\n|$)))",
            # postal address with only street name and number :  EX : 9/11 rue Jeanne d'Arc
            r"^.{,3}\d+(?:[ /-]?\d+)? " + street_word_pattern + r".{,50}$",
            # EX number 2 : 23, rue de la Monnaie
            r"^.{,5}" + street_word_pattern + r".{,50}$",
            # postal address with only postal code and city name  :   EX : 76000 ROUEN
            r"^.{,3}\d{5}[\xa0| ][A-Z].{,5}$",
            # Known job title
            line_with_known_job,
            # Contact
            r"^.{,15}Pour nous contacter.{,15}$",
            r"^.{,3}Contact (e.mail|t[eé]l[eé]phone).{,3}$",
            # email adress EX: Adresse mail : cyrimmmman80@gmail.com
            r"^.{,3}([Aa]dresse mail|Mail).{,3}" + email_address_regex + r"$",
            # address with date EX : Torroella de Montgri, le 5 avril 2023
            r"^[A-Za-z]+(?: [A-Za-z]+)*, le \d{1,2} [A-Za-z]+ \d{4}.{,3}$",
        ]

    def post_process_tags(self, tags: list[tuple[str, str]]) -> list[tuple[str, str]]:
        """
        Method to apply tagging rules posterior to the standard regex tagging.

        Parameters
        ----------
        tags: list[tuple[str, str]]
            Original tags

        Returns
        -------
        _: list[tuple[str, str]]
            Post-processed tags
        """
        # Signature lines containing first/last name
        tags = self.detect_name_signature(tags)

        return tags

    def detect_name_signature(self, tags: list[tuple[str, str]]) -> list[tuple[str, str]]:
        """
        Method to detect lines containing First name / Surname
        Ex: Mr Joe Dupond

        Parameters
        ----------
        tags: list[tuple[str, str]]
            Original tags

        Returns
        -------
        _: list[tuple[str, str]]
            Post processed tags
        """
        # First name / Last name Signatures
        capitalized_words: str = r"[A-Z][-'A-za-zÀ-ÿ]{,10}"
        particles: str = r"le|d[ei]|d'?|v[oa]n|del"
        line_with_name: str = (
            rf"(?:^[ >]*{capitalized_words}(?:-{capitalized_words})?(?:(?: +(?:{particles}"
            rf"))?\.? {{,2}}{capitalized_words}(?:-{capitalized_words})?){{1,4}}\.? *(?:\n+|$))"
        )

        # Forbidden words (lowercase)
        forbidden_words: set[str] = {"urgent", "attention"}

        new_tags: list[tuple[str, str]] = list()
        for tag, text in tags:
            if tag == self.default_tag:
                match = re.match(line_with_name, text)
                has_forbidden_words: bool = bool(forbidden_words.intersection(text.lower().split()))

                if match and not has_forbidden_words:
                    tag = "SIGNATURE_NAME"

            new_tags.append((tag, text))

        return new_tags


class TransferredEmailProcessor(MelusineTransformer):
    """
    Processing specific to transferred emails such as:
    - Extracting the email address of the original sender (before transfer)
    - Removing empty messages related to the transfer action
    """

    def __init__(
        self,
        output_columns: Iterable[str] = ("messages", "det_clean_from"),
        tags_to_ignore: Iterable[str] = ("FOOTER", "SIGNATURE"),
        messages_column: str = "messages",
    ):
        """
        Parameters
        ----------
        output_columns: str
        tags_to_ignore: Iterable[str]
            If a message contains only tags in this list, it will be ignored
        messages_column: DataFrame column containing a list of Message instances
        """
        self.messages_column = messages_column
        input_columns = [self.messages_column]

        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
            func=self.process_transfered_mail,
        )

        self.tags_to_ignore = tuple(tags_to_ignore)
        self.json_exclude_list.append("input_columns")

    @property
    def email_pattern(self) -> str:
        """
        Regex pattern to detect an email address.
        """
        return r"\w+(?:[-+.']\w+)*@\w+(?:[-.]\w+)*\.\w+(?:[-.]\w+)*"

    @property
    def meta_email_address_regex(self) -> str:
        """
        Regex to extract an email address from an email transition pattern.
        Ex:
            De: jane@gmail.fr A: joe@gmail.fr Envoyé à: 11h22
            => jane@gmail.fr
        """
        # De: jane@gmail.fr A: joe@gmail.fr Envoyé à: 11h22
        # De:\nJane <jane@gmail.fr>\nA: joe@gmail.fr Envoyé à: 11h22
        # De: Jane "jane@gmail.fr"\nA: joe@gmail.fr Envoyé à: 11h22
        start_pattern = r"(?:Message de\b|\bDe\b|Exp[ée]diteur|\bFrom\b) ?\n? ?: *\n?(?:.{,30}[ <\"])?"
        email_pattern = self.email_pattern
        end_pattern = r"(?: *<.*>)?(?:.{,5}(?:$|\n)|.{,5}(?:Envoy|A|Sent|À))"
        meta_pattern_1 = f"{start_pattern}({email_pattern}){end_pattern}"

        # Le 28 févr. 2023 à 10:33, joee@gmail.fr a écrit :
        # Le dim., 28 févr. 2023 à 10:33, joee@gmail.fr a écrit :
        # Le 01/01/2001 10:33, Joe <joee@gmail.fr> a écrit :
        start_pattern = r"Le (?:\d.{,15}[aà]|[a-z]{3}.{,50}[aà]|\d{2}/\d{2}/\d{4}) \d{,2}[:hH]\d{,2}.{,30} <?"
        end_pattern = r">? a [ée]crit ?:"
        meta_pattern_2 = f"{start_pattern}({email_pattern}){end_pattern}"

        # Assemble patterns
        meta_pattern = "|".join([meta_pattern_1, meta_pattern_2])

        return meta_pattern

    def process_transfered_mail(self, message_list: list[Message]) -> tuple[list[Message], str | None]:
        """
        Run all transformations related to transfer emails.

        Args:
            message_list: Emails input data

        Returns:
            message_list: List of messages in the conversation
            clean_address_from: Processed sender email address
        """
        clean_address_from: str | None = None

        # Filter out transfer message (contains only irrelevant tags)
        message_list = self.filter_message_list(message_list)

        # Extract email address data from transition pattern
        top_message = message_list[0]
        extracted_address_from: str | None = self.extract_email_address(top_message)

        # If no address
        if extracted_address_from:
            clean_address_from = extracted_address_from

        return message_list, clean_address_from

    def extract_email_address(self, message: Message) -> str | None:
        """
        Extract sender email address from message meta (transition pattern).

        Args:
            message: Message with text and metadata

        Returns:
            extracted_address_from: Extracted sender address if available
        """
        extracted_address_from = None

        if message.meta:
            # Extract email address
            match_list = re.findall(self.meta_email_address_regex, message.meta)

            # Filter out empty matches
            match_list = [match for match_group in match_list for match in match_group if match]

            # Sanity check on address
            if match_list and "@" in match_list[0]:
                extracted_address_from = match_list[0]

        return extracted_address_from

    def filter_message_list(self, message_list: list[Message]) -> list[Message]:
        """ """
        top_message = message_list[0]

        parts = top_message.extract_parts()
        contains_only_tags_to_ignore = all([tag.startswith(self.tags_to_ignore) for tag, _ in parts])

        if contains_only_tags_to_ignore and (len(message_list) > 1):
            message_list = message_list[1:]

        return message_list


class DeterministicTextFlagger(MelusineTransformer):
    """
    Class to flag text patterns such as :
    "new york" => "new_york"
    """

    def __init__(
        self,
        text_flags: dict[str, Any],
        input_columns: str = "text",
        output_columns: str = "text",
        remove_multiple_spaces: bool = True,
        add_spaces: bool = True,
    ):
        """
        Parameters
        ----------
        text_flags: dict[str, Any]
            Dict containing flag name as key and regex pattern as value
        add_spaces: bool
            If true, add spaces around flags
        remove_multiple_spaces: bool
            If True, remove multiple spaces after flagging
        """
        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
            func=self.flag_text,
        )
        self.text_flags = text_flags
        self.add_spaces = add_spaces
        self.remove_multiple_spaces = remove_multiple_spaces

    @staticmethod
    def default_flag_text(
        text: str,
        flag_dict: dict[str, str],
        add_spaces: bool = True,
        remove_multiple_spaces: bool = True,
    ) -> str:
        """
        Method to apply flagging on a text.
        General flagging: replace remarkable expressions by a flag
        Ex: 0123456789 => flag_phone_

        Parameters
        ----------
        flag_dict: dict[str, str]
            Flagging dict with regex as key and replace_text as value
        text: str
            Text to be flagged
        add_spaces: bool
            If true, add spaces around flags
        remove_multiple_spaces: bool
            If True, remove multiple spaces after flagging

        Returns
        -------
        text: str
            Flagged text
        """
        # Support for nested flag dicts
        for key, value in flag_dict.items():
            if isinstance(value, dict):
                text = DeterministicTextFlagger.default_flag_text(
                    text=text,
                    flag_dict=value,
                    add_spaces=add_spaces,
                    remove_multiple_spaces=remove_multiple_spaces,
                )
            else:
                # Add spaces to avoid merging words with flags
                if add_spaces:
                    replace_value = " " + value + " "
                else:
                    replace_value = value
                text = re.sub(key, replace_value, text, flags=re.I)

            if remove_multiple_spaces:
                text = re.sub(" +", " ", text)
        return text.strip()

    def flag_text(self, text: str) -> str:
        """
         Method to flag text.

         Parameters
         ----------
         text: str
             Text to be flagged

         Returns
         -------
        _: str
             Flagged text
        """
        # Join collocations
        text = self.default_flag_text(
            text,
            self.text_flags,
            add_spaces=self.add_spaces,
            remove_multiple_spaces=self.remove_multiple_spaces,
        )

        return text


class Cleaner(MelusineTransformer):
    """
    Class to clean text columns
    """

    def __init__(
        self,
        substitutions: dict[str, Any],
        input_columns: str = "text",
        output_columns: str = "text",
    ):
        """
        Parameters
        ----------
        substitutions: dict[str, Any]
            Dict containing replace pattern and replacement value
        """
        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
            func=self.clean,
        )
        self.substitutions = substitutions

    def clean(self, text: str) -> str:
        """
         Method to clean text.

         Parameters
         ----------
         text: str
             Text to be flagged

         Returns
         -------
        _: str
             Flagged text
        """
        # Join collocations
        text = DeterministicTextFlagger.default_flag_text(
            text,
            self.substitutions,
            add_spaces=False,
            remove_multiple_spaces=True,
        )

        return text


class DateProcessor(MelusineTransformer):

    """
    Parse string date to iso format string date
    """

    ISO_FORMAT = "%Y-%m-%d"
    LANGUAGE = ["en_US", "fr_FR", "es_ES", "it_IT", "nl_NL", "de_DE", "tr_TR"]
    CALENDAR_ABBR = {
        "lu,": "lun",
        "ma,": "mar",
        "me,": "mer",
        "je,": "jeu",
        "jeudii": "jeudi",
        "ve,": "ven",
        "sa,": "sam",
        r"^di,": "dim",
        r"\sjune\s": " jun ",
        r"\sjuly\s": " jul ",
    }
    FRENCH_CALENDAR_ABBR = {
        "fev ": "févr ",
        "jun": "juin ",
        "jul ": "juil ",
        "aout": "août",
        "sep ": "sept ",
        "dec ": "déc ",
        "iuillet": "juillet",
        "juihet": "juillet",
        "sptembre": "septembre",
        "seplentre": "septembre",
        "septembm": "septembre",
        "septernbre": "septembre",
        "juillat": "juillet",
        "decembre": "décembre",
        "fevrier": "février",
    }
    PATTERN_DATE = {
        r"(\d{4}-\d{2}-\d{2})": "YYYY-MM-DD",
        r"(\d{4}/\d{2}/\d{2})": "YYYY/MM/DD",
        r"(\d{2}-\d{2}-\d{4})": "DD-MM-YYYY",
        r"(\d{2}/\d{2}/\d{4})": "DD/MM/YYYY",
        r"([A-zÀ-ÿ]{4,10}, \d{1,2} [A-zÀ-ÿ]{3,4}, \d{4})": "dddd, DD MMM, YYYY",
        r"([A-zÀ-ÿ]{4,10}, \d{1,2} [A-zÀ-ÿ]{3,4} \d{4})": "dddd, DD MMM YYYY",
        r"([A-zÀ-ÿ]{4,10} \d{1,2} [A-zÀ-ÿ]{3,4} \d{4})": "dddd DD MMM YYYY",
        r"([A-zÀ-ÿ]{4,10}, \d{1,2} [A-zÀ-ÿ]{,10}, \d{4})": "dddd, DD MMMM, YYYY",
        r"([A-zÀ-ÿ]{4,10}, \d{1,2} [A-zÀ-ÿ]{,10} \d{4})": "dddd, DD MMMM YYYY",
        r"([A-zÀ-ÿ]{4,10} \d{1,2} [A-zÀ-ÿ]{,10} \d{4})": "dddd DD MMMM YYYY",
        r"(\d{1,2} [A-zÀ-ÿ]{3,4} \d{4})": "DD MMM YYYY",
        r"(\d{1,2} [A-zÀ-ÿ]{3,4}, \d{4})": "DD MMM, YYYY",
        r"(\d{1,2} [A-zÀ-ÿ]{,10} \d{4})": "DD MMMM YYYY",
        r"(\d{1,2} [A-zÀ-ÿ]{,10}, \d{4})": "DD MMMM, YYYY",
        r"([A-zÀ-ÿ]{3}, \d{1,2} [A-zÀ-ÿ]{3}, \d{4})": "ddd, DD MMM, YYYY",
        r"([A-zÀ-ÿ]{3}, \d{1,2} [A-zÀ-ÿ]{3} \d{4})": "ddd, DD MMM YYYY",
        r"([A-zÀ-ÿ]{3}\. \d{1,2} [A-zÀ-ÿ]{,10}\.? \d{4})": "ddd DD MMM YYYY",
        r"([A-zÀ-ÿ]{2}, \d{1,2} [A-zÀ-ÿ]{3}, \d{4})": "ddd, DD MMM, YYYY",
        r"([A-zÀ-ÿ]{2}, \d{1,2} [A-zÀ-ÿ]{3} \d{4})": "ddd, DD MMM YYYY",
        r"(\d{1,2} [A-zÀ-ÿ]{3} \d{4})": "DD MMM YYYY",
    }

    def __init__(
        self,
        input_columns: str = "date",
        output_columns: str = "date",
    ) -> None:
        """
        Parameters
        ----------
        input_columns: str
            Input columns for the transform operation
        output_columns: str
            Outputs columns for the transform operation
        """
        super().__init__(
            input_columns=input_columns,
            output_columns=output_columns,
            func=self.parse_date_to_iso,
        )

    @classmethod
    def parse_date_to_iso(cls, date_: str) -> str | None:
        """
        This function use the package arrow to convert a date from string format with any
        type of format (i.e. vendredi 8 juillet 2020 -> 2020-07-08)
        This package is prefered to datetime because datetime use locale settings and
        arrow do not use locale setting
        Visit https://arrow.readthedocs.io/en/latest/ for more information

        Parameters
        ----------
        date_: str
                date read from mail with any format

        Returns
        -------
        date_: str
                date_ as string with iso format (YYYY-MM-DD)
        """
        # Initialization
        matched_group: str | None = None
        date_ = date_ or ""
        date_ = date_.lower()

        for pattern, format_ in cls.PATTERN_DATE.items():
            pattern_compiled = re.compile(pattern)
            matched_date_format = pattern_compiled.search(date_)
            if matched_date_format is not None:
                matched_group = matched_date_format.group()

                # Replace single digit by 0+digit (i.e. 9 -> 09)
                matched_group = cls.process_single_digit(matched_group, pattern)

                # Replace some known abbreviations
                for key, value in cls.CALENDAR_ABBR.items():
                    matched_group = re.sub(key, value, matched_group)

                matched_group = cls.convert_to_iso_format(matched_group, format_)

                # The format matched so we end the for loop
                break

        return matched_group

    @staticmethod
    def process_single_digit(matched_group: str, pattern: str) -> str:
        """
        Replace single digit by 0+digit
        i.e.: 9 -> 09
        """
        if r"\d{1,2}" in pattern:
            # Case when digit is in the middle of the string
            number_searched = re.search(r"\s\d\s", matched_group)
            if number_searched is not None:
                number = number_searched.group().strip()
                matched_group = matched_group.replace(f" {number} ", f" 0{number} ")
            # Case when digit is the first char of the string
            number_searched = re.search(r"^\d\s", matched_group)
            if number_searched is not None:
                number = number_searched.group().strip()
                matched_group = re.sub(r"^\d\s", f"0{number} ", matched_group)
        return matched_group

    @classmethod
    def convert_to_iso_format(cls, matched_group: str, format_: str) -> str | None:
        """
        Try to convert the date found as any string form to ISO format
        """
        # In case we are working with abbreviations
        if "ddd" in format_ or "MMM" in format_:
            for lang in cls.LANGUAGE:
                try:
                    matched_group_copy = matched_group
                    if lang == "fr_FR":
                        matched_group_copy = re.sub("[^A-zÀ-ÿ0-9:, ]+", "", matched_group_copy)
                        for key, value in cls.FRENCH_CALENDAR_ABBR.items():
                            matched_group_copy = re.sub(key, value, matched_group_copy)
                    matched_group = arrow.get(matched_group_copy, format_, locale=lang).datetime.strftime(
                        cls.ISO_FORMAT
                    )
                    break
                except arrow.parser.ParserMatchError:
                    pass

            # ISO Format
            if re.search(r"\d{4}-\d{2}-\d{2}", matched_group):
                return matched_group

            # We failed finding the abbreviation so we give up and return None
            return None

        # We were not working with abbrevations so we use arrow package
        return arrow.get(matched_group, format_).datetime.strftime(cls.ISO_FORMAT)



from typing import Dict, List, Optional, Union

from melusine.base import MelusineRegex


class VacationReplyRegex(MelusineRegex):
    """
    Detect vacation reply patterns such as "Je suis absent du bureau".
    """

    @property
    def positive(self) -> Union[str, Dict[str, str]]:
        """
        Define regex patterns required to activate the MelusineRegex.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return dict(
            VAC_REP_URGENCY=r"en\s+?cas\s+?d'?\s?urgenc.{1,100}(?:contact|app?eler?)|pour toute urgence.{1,100}contact",
            VAC_REP_HOLIDAYS=(
                r"^.{,30}(?:je suis |etant )?(:?actuellement)?(?:(?<!...\bete|.\betai[st]|etions|.\betiez)\sabsente?|"
                r"(?<!...\bete|.\betai[st]|etions|.\betiez)\sen cong[ée]s?)"
            ),
            VAC_REP_ON_MOVE=(
                r"(?<!serai\s|etais\s)en deplacement[_\s]prof|^.{1,15}(?<!...\bete|.\betai[st]|etions|.\betiez)"
                r"\sabsente?.{1,15}du.{1,10}flag_date_|^(?:absente?|en cong[ée]s?)"
            ),
            VAC_REP_OUT_OF_OFFICE=r"je prendra.{,7}connaissance",
            VAC_REP_OUT_OF_OFFICE_ENG=r"^.{,35}(?:i\s(?:am|will be)\s)?(:?currently\s)?(?:out of (?:the )?office)",
            VAC_REP_NO_REPLY=r"(r[eé]?ponse automatique|automatic reply|noreply)",
            VAC_REP_AUTO=r"cec[il] est (?:une r[ée]ponse|un (e-?)?mail) automatique",
            VAC_REP_ACKNOLEDGMENT=r"(?:^|par cet? e?mail,?\s)Nous\saccusons?\s(bonne\s)?r[ée]ception",
        )

    @property
    def neutral(self) -> Optional[Union[str, Dict[str, str]]]:
        """
        Define regex patterns to be ignored when running detection.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return None

    @property
    def negative(self) -> Optional[Union[str, Dict[str, str]]]:
        """
        Define regex patterns prohibited to activate the MelusineRegex.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return dict(
            VAC_REP_FORBIDDEN_PROC=r"pas.{,10}traite|pas.{,10}identifie|pas.{,10}aboutir?",
            VAC_PREP_FORBIDDEN_ACTION=(
                r"dans.{,7}attente|merci par avance|(?:serai[st]?|aurai[st]?) souhait(able)?|relanc[ée]r?"
            ),
            VAC_PREP_FORBIDDEN_FORGOT=(
                r"je.{,10}perm(?:ets?|is?)|effectivement|\boublie?|malheureusement|suite [aà]|"
                r"(?:monsieur|madame).{,15}absent?"
            ),
            VAC_REP_FORBIDDEN_RETURN=(
                r"(?:retour|reponse) tardi(?:f|ve)|vien[ts] de (?:prendre (?:note|conn?aiss?ance)|recevoir)"
            ),
            VAC_REP_FORBIDDEN_EXTRA=(
                r"\bai\b.{,15}app?ell?er?|\bai\b.{,7}\blu[es]?\b|\barr?eter?.{,5}date|(?:aucune|pas).{,7}nouvelle|"
                r"chaque ann[ée]e|voisin"
            ),
            VAC_REP_FORBIDDEN_AMBIGUOUS=r"en\s+?cas\s+?de\s?besoin.{1,100}(?:contact|app?eler?)|going on leave",
            VAC_REP_FORBIDDEN_ABSENCE=(
                r"dur[ée]e? ind[ée]termin[ée]e?|"
                r"expertise judiciaire|doi.{,35}au plus tard|^.{,50}adresse[rz] vos|en mon absence"
            ),
            VAC_REP_FORBIDDEN_EMAIL=(
                r"pas prise? en compte|chang[ée]r?.{,5}(?:adresse|mail)|"
                r"mail.{,15}(?:(in)?acti|pas lu)|acc[eè]s limit[ée]"
            ),
            VAC_REP_FORBIDDEN_SICKNESS=r"arr?[eê]t de travail|\bmaternit[ée]\b|maladie",
            VAC_REP_FORBIDDEN_LEAVE=(
                r"\bposte\b.{,50}\bvacant\b|d[ée]finitivement|succe[ée]d[ée]|"
                r"quitt[ée]|remplacante?|plus parti.{,25}cabinet|retraite"
            ),
            VAC_REP_FORBIDDEN_HOME=(
                r"^.{,15}(?:devant ).{,5}(?:absente|en cong[ée]s?)|"
                r"^.{,15}absente? de (?:mon domicile?|chez|.{,5}maison)"
            ),
            VAC_REP_FORBIDDEN_CO=r"entreprise|document",
        )

    @property
    def match_list(self) -> List[str]:
        """
        List of texts that should activate the MelusineRegex.

        Returns:
            _: List of texts.
        """
        return [
            "en cas d'urgence, vous pouvez m'appeler sur mon mobile au 01 02 03 04 05",
            "pour toute urgence, vous pouvez me contacter au 01 02 03 04 05",
            "étant actuellement absent, je vous répondrais lors de mon retour le 01/01/2001",
            "Je suis en congé, je vous répondrais à mon retour",
            "Je suis absent du flag_date au flag_date",
            "je prendrais connaissance de votre message dès mon retour",
            "I am currently out of office",
            "ceci est une réponse automatique",
        ]

    @property
    def no_match_list(self) -> List[str]:
        """
        List of texts that should NOT activate the MelusineRegex.

        Returns:
            _: List of texts.
        """
        return [
            "Je souhaite une réponse même si je suis actuellement en congé",
            "en attendant, je suis en congé, dans l'attente de votre réponse",
            "veuillez m'excuser, j'ai oublié de répondre avant de partir en vacances",
            "excusez mon retour tardif, j'étais en déplacement professionnel",
            "je suis absent pour une durée indéterminée, veuillez contacter Jane",
            "je suis actuellement en congé avec un accès limité à ma messagerie",
            "je suis en congé maladie",
            "je suis en congé maternité",
            "Je suis absent car je suis parti en retraite",
        ]



"""
Unit test of the ThanksDetector.

"""
from tempfile import TemporaryDirectory

import pandas as pd
import pytest

from melusine.base import MissingFieldError
from melusine.detectors import ThanksDetector
from melusine.message import Message


@pytest.fixture
def thanks_detector_df():
    m0 = Message("")
    m0.tags = [
        ("HELLO", "Bonjour"),
        ("THANKS", "Merci beaucoup"),
    ]
    m0_messages = [m0]
    m0_expected = True
    m0_debug_expected = {
        "match_result": True,
        "negative_match_data": {},
        "neutral_match_data": {},
        "positive_match_data": {"DEFAULT": [{"match_text": "Merci", "start": 0, "stop": 5}]},
    }

    m1 = Message("")
    m1.tags = [
        ("HELLO", "Bonjour"),
        ("THANKS", "Merci, j'attends une reponse"),
    ]
    m1_messages = [m1]
    m1_expected = False
    m1_debug_expected = {
        "match_result": False,
        "negative_match_data": {"FORBIDDEN_WORDS": [{"match_text": "attend", "start": 9, "stop": 15}]},
        "neutral_match_data": {},
        "positive_match_data": {"DEFAULT": [{"match_text": "Merci", "start": 0, "stop": 5}]},
    }

    df = pd.DataFrame(
        {
            "messages": [m0_messages, m1_messages],
            "detection_expectation": [m0_expected, m1_expected],
            "debug_expectation": [m0_debug_expected, m1_debug_expected],
        }
    )

    return df


def test_thanks_detector(thanks_detector_df):
    """Unit test of the debug mode."""
    df = thanks_detector_df
    df_copy = df.copy()
    detector = ThanksDetector(
        name="thanks",
    )
    result_col = detector.result_column
    debug_dict_col = detector.debug_dict_col

    # Transform data
    df.debug = True
    df = detector.transform(df)

    # Test result
    assert result_col in df.columns
    assert debug_dict_col in df.columns

    for i, row in df.iterrows():
        assert row[result_col] == row["detection_expectation"]
        assert row[debug_dict_col][detector.thanks_regex.regex_name] == row["debug_expectation"]


def test_thanks_detector_missing_field(thanks_detector_df):
    """Unit test of the debug mode."""
    df = thanks_detector_df.copy()

    detector = ThanksDetector(
        name="thanks",
    )
    df = df.drop(detector.input_columns, axis=1)

    # Transform data
    with pytest.raises(MissingFieldError, match=str(detector.input_columns)):
        _ = detector.transform(df)


@pytest.mark.parametrize(
    "tags, has_body, thanks_text, thanks_parts",
    [
        (
            [
                ("HELLO", "Bonjour madame"),
                ("BODY", "Voici le dossier"),
                ("THANKS", "Merci a vous"),
            ],
            True,
            "Merci a vous",
            [("THANKS", "Merci a vous")],
        ),
        (
            [
                ("HELLO", "Bonjour madame"),
                ("THANKS", "Merci"),
                ("THANKS", "Merci a vous"),
            ],
            False,
            "Merci\nMerci a vous",
            [("THANKS", "Merci"), ("THANKS", "Merci a vous")],
        ),
    ],
)
@pytest.mark.usefixtures("use_dict_backend")
def test_thanks_detector_debug(tags, has_body, thanks_text, thanks_parts):
    """Unit test of the debug mode."""

    data = {
        "messages": [Message(text="", tags=tags)],
        "debug": True,
    }

    detector = ThanksDetector(
        name="thanks",
    )

    # Transform data
    data = detector.transform(data)

    # Test result
    assert "debug_thanks" in data
    assert "has_body" in data["debug_thanks"]
    assert "thanks_text" in data["debug_thanks"]
    assert "thanks_parts" in data["debug_thanks"]

    assert data["debug_thanks"]["has_body"] == has_body
    assert data["debug_thanks"]["thanks_text"] == thanks_text
    assert data["debug_thanks"]["thanks_parts"] == thanks_parts



"""
Melusine transformation can operate on different data structures such as dict or pandas.DataFrame.
Different transformation backends are used to process different data structures.
The BaseTransformerBackend class defines the interface for transformation backend classes.

Implemented classes: [
    BaseTransformerBackend,
]
"""
from abc import ABC, abstractmethod
from typing import Any, Callable, List, Optional


class BaseTransformerBackend(ABC):
    """
    Abstract base class defining how to implement a Melusine Backend.
    Each backend applies transform operations on a specific type of data.
    Ex: Pandas DataFrames, Dict, Spark objects, etc
    """

    DEBUG_FLAG = "debug"

    @abstractmethod
    def apply_transform(
        self,
        data: Any,
        func: Callable,
        output_columns: Optional[List[str]] = None,
        input_columns: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Method to apply a transform on a Dataset using the current backend.

        Parameters
        ----------
        data: Any
            Data to be transformed
        func: Callable
            Transform function to apply to the input data
        output_columns: List[str]
            List of output columns
        input_columns: List[str]
            List of input columns
        kwargs

        Returns
        -------
        _: Any
            Transformed data
        """

    @abstractmethod
    def add_fields(self, left: Any, right: Any, fields: Optional[List[str]] = None) -> Any:
        """
        Method to add fields form the right object to the left object.

        Parameters
        ----------
        left: Dataset
            MelusineDataset object
        right: Dataset
            Melusine Dataset object
        fields: List[str]
            List of fields to be added

        Returns
        -------
        _: Dataset
            Left object with added fields
        """

    @abstractmethod
    def copy(self, data: Any, fields: Optional[List[str]] = None) -> Any:
        """
        Method to make a copy of the dataset.

        Parameters
        ----------
        data: Dataset
            MelusineDataset object
        fields: List[str]
            List of fields to include in the copy (by default copy all fields)

        Returns
        -------
        _: Dataset
            Copy of original object
        """

    @abstractmethod
    def get_fields(self, data: Any) -> List[str]:
        """
        Method to get the list of fields available in the input dataset.

        Parameters
        ----------
        data: Dataset
            MelusineDataset object

        Returns
        -------
        _: List[str]
            List of dataset fields
        """

    def check_debug_flag(self, data: Any) -> bool:
        """
        Method to check if debug_mode is activated.

        Parameters
        ----------
        data: Dataset
            MelusineDataset object

        Returns
        -------
        _: bool
            True if debug mode is activated
        """
        if hasattr(data, self.DEBUG_FLAG):
            debug_mode = getattr(data, self.DEBUG_FLAG)
        elif isinstance(data, dict):
            debug_mode = data.get(self.DEBUG_FLAG, False)
        else:  # pragma: no cover
            debug_mode = False

        return debug_mode

    @abstractmethod
    def setup_debug_dict(self, data: Any, dict_name: str) -> Any:
        """
        Method to check if debug_mode is activated.

        Parameters
        ----------
        data: Dataset
            MelusineDataset object
        dict_name: str
            Name of the debug dict field to be added

        Returns
        -------
        _: Dataset
            MelusineDataset object
        """



import re

from melusine import config
from melusine.message import Message


def test_message_repr():
    message = Message(text="Hello")

    assert re.search(r"meta='NA'", repr(message))
    assert re.search(r"text='Hello'", repr(message))

    message = Message(text="Hello", meta="someone@domain.fr")

    assert re.search(r"meta='someone@domain.fr'", repr(message))
    assert re.search(r"text='Hello'", repr(message))


def test_message_has_tags():
    message = Message(text="Hello")
    message.tags = [
        ("HELLO", "Bonjour"),
        ("BODY", "Pouvez-vous"),
        ("GREETINGS", "Cordialement"),
    ]

    assert not message.has_tags(target_tags=["FOOTER"])
    assert message.has_tags(target_tags=["BODY"])
    assert message.has_tags(target_tags=["FOOTER", "HELLO"])


def test_message_has_tags_stop_at():
    message = Message(text="Hello")
    message.tags = [
        ("HELLO", "Bonjour"),
        ("GREETINGS", "Cordialement"),
        ("BODY", "Blah Blah Blah"),
    ]

    assert not message.has_tags(target_tags=["BODY"], stop_at=["GREETINGS"])


def test_message_has_tags_no_tags():
    message = Message(text="Hello")

    assert not message.has_tags(target_tags=["BODY"])


def test_message_extract_parts():
    message = Message(text="Hello")
    message.tags = [
        ("HELLO", "Bonjour"),
        ("BODY", "Pouvez-vous"),
        ("GREETINGS", "Cordialement"),
    ]

    assert message.extract_parts(target_tags={"BODY"}) == [("BODY", "Pouvez-vous")]
    assert message.extract_parts(target_tags=["GREETINGS", "HELLO"]) == [
        ("HELLO", "Bonjour"),
        ("GREETINGS", "Cordialement"),
    ]


def test_message_extract_parts_stop():
    message = Message(text="Hello")
    message.tags = [
        ("HELLO", "Bonjour"),
        ("FOOTER", "Envoyé depuis mon Iphone"),
        ("GREETINGS", "Cordialement"),
        ("BODY", "Blah Blah Blah"),
    ]

    extracted = message.extract_parts(target_tags=["BODY"], stop_at=["FOOTER", "GREETINGS"])

    assert extracted == []


def test_message_extract_parts_no_tags():
    message = Message(text="Hello")

    assert not message.extract_parts(target_tags={"BODY"})


def test_message_extract_last_body():
    message = Message(text="Hello")
    message.tags = [
        ("HELLO", "Bonjour"),
        ("BODY", "Pouvez-vous"),
        ("GREETINGS", "Cordialement"),
    ]

    assert message.extract_last_body() == [("BODY", "Pouvez-vous")]


def test_str():
    # Arrange
    message = Message(meta="Test\nmeta", text="Hello")
    message.tags = [
        ("TAG", "ABC"),
        ("TAAG", "ABCD"),
        ("TAAAG", "ABCDE"),
    ]

    expected_list = [
        r"=+ +Message +=+",
        r"-+ +Meta +-+",
        r"Test",
        r"meta",
        r"-+ +Text +-+",
        r"ABC\.+TAG",
        r"ABCD\.+TAAG",
        r"ABCDE\.+TAAAG",
        r"=+",
    ]

    # Act
    result = str(message).strip()

    # Assert
    assert len(result.splitlines()) == len(expected_list)
    for text_line, regex in zip(result.splitlines(), expected_list):
        assert re.match(regex, text_line)


def test_str_no_meta():
    # Arrange
    message = Message(text="Hello")
    message.tags = [
        ("TAG", "ABC"),
        ("TAAG", "ABCD"),
        ("TAAAG", "ABCDE"),
    ]

    expected_list = [
        r"=+ +Message +=+",
        r"-+ +Meta +-+",
        r"N/A",
        r"-+ +Text +-+",
        r"ABC\.+TAG",
        r"ABCD\.+TAAG",
        r"ABCDE\.+TAAAG",
        r"=+",
    ]

    # Act
    result = str(message).strip()

    # Assert
    assert len(result.splitlines()) == len(expected_list)
    for text_line, regex in zip(result.splitlines(), expected_list):
        assert re.match(regex, text_line)


def test_str_no_tags():
    # Arrange
    message = Message(text="Hello")

    expected_list = [
        r"=+ +Message +=+",
        r"-+ +Meta +-+",
        r"N/A",
        r"-+ +Text +-+",
        r"Hello",
        r"=+",
    ]

    # Act
    result = str(message).strip()

    # Assert
    assert len(result.splitlines()) == len(expected_list)
    for text_line, regex in zip(result.splitlines(), expected_list):
        assert re.match(regex, text_line)


def test_str_no_conf(reset_melusine_config):
    config.reset({"Test": "Test"})
    message = Message(text="test", tags=[("TEST TAG", "TEST TEXT")])
    print(message)
    assert True



"""
Module that contains utility functions for tests (in /tests).
"""
from typing import Any, Dict

from melusine.base import MelusineTransformer
from melusine.pipeline import MelusinePipeline

expected_suffix: str = "_expected"


def assert_pipeline_results(email: Dict[str, Any], pipeline_name: str) -> None:
    """
    Assert that the pipeline execution result correspond to the testcase expectation.

    Parameters
    ----------
    email: Dict[str, Any]
        Email content.
    pipeline_name: str
        Name of a Melusine pipeline.
    """
    # Instantiate Pipeline
    pipeline: MelusinePipeline = MelusinePipeline.from_config(pipeline_name)

    # Useful fields
    test_name: str = email["test_name"]

    # Loop on pipeline transformers
    for transformer_name, transformer in pipeline.steps:
        email = assert_transformation(email, transformer, transformer_name, test_name)

    # Check that the pipeline returns a dict
    assert isinstance(email, dict)

    # Look for untested fields
    untested_fields = [x for x in email if x.endswith(expected_suffix)]
    assert not untested_fields, f"Field(s) {untested_fields} have not been tested for" f"Test-case '{test_name}'"


def assert_transformation(
    email: Dict[str, Any], transformer: MelusineTransformer, transformer_name: str, test_name: str
) -> Dict[str, Any]:
    """

    Parameters
    ----------
    email: Dict[str, Any]
        Email data dict
    test_name: str
        Name of the current test
    transformer: MelusineTransformer
        Data transformer instance
    transformer_name: str
        Name of the current transformer

    Returns
    -------
    email: Dict[str, Any]
        Transformed email data dict
    """
    # Apply transformer on email
    email = transformer.transform(email)

    # Do we have an expected value for this transformer?
    expected_key = f"{transformer_name}{expected_suffix}"
    if expected_key in email:
        expectation_dict: Dict[str, Any] = email.pop(expected_key)

        # Loop on columns with an expected value
        for col, expected_value in expectation_dict.items():
            # Specific case for the message column
            if col.startswith("messages"):
                assert_message_attribute(col, email, expected_value, test_name, transformer_name)

            # Regular case
            else:
                assert email[col] == expected_value, (
                    f"Failure for test {test_name} at step {transformer_name}. "
                    f"Value expected for column {col} : {expected_value}. "
                    f"Value obtained for column {col} : {email[col]}"
                )
    return email


def assert_message_attribute(
    col: str, email: Dict[str, Any], expected_value: Any, test_name: str, transformer_name: str
) -> None:
    """

    Parameters
    ----------
    col: str
        Column name
    email: Dict[str, Any]
        Email data dict
    expected_value: Any
        Expected message attribute value
    test_name: str
        Name of the current test
    transformer_name: str
        Name of the current transformer

    Returns
    -------

    """
    # Check the number of messages in the conversation
    assert len(email["messages"]) == len(expected_value)

    # Loop on messages
    for message, expected_attr in zip(email["messages"], expected_value):
        _, attribute_name = col.rsplit(".")
        attribute_value = getattr(message, attribute_name)

        # Test attribute values
        assert attribute_value == expected_attr, (
            f"Failure for test {test_name} at step {transformer_name}. "
            f"Value expected for attribute {attribute_name} : {expected_attr}. "
            f"Value obtained for attribute {attribute_name} : {attribute_value}"
        )



"""
Top-level package.
"""
import logging
from ctypes import CDLL, cdll
from typing import Any, Optional

import pandas as pd

from melusine._config import config

__all__ = ["config"]

VERSION = (3, 0, 0)
__version__ = ".".join(map(str, VERSION))

# ------------------------------- #
#             LOGGING
# ------------------------------- #
logging.getLogger(__name__).addHandler(logging.NullHandler())


# ------------------------------- #
#           MONKEY PATCH
# ------------------------------- #

# Monkey patch for pandas DataFrame memory leaking on linux OS (pandas issue #2659)
try:
    # Try executing linux malloc_trim function (release free memory)
    cdll.LoadLibrary("libc.so.6")
    libc: Optional[CDLL] = CDLL("libc.so.6")
    if libc is not None:
        libc.malloc_trim(0)
except (OSError, AttributeError):  # pragma: no cover
    # Incompatible OS: this monkey patch is not needed
    libc = None

# Store the standard pandas method
__std_del: Optional[Any] = getattr(pd.DataFrame, "__del__", None)


# Prepare a new __del__ method
def __fixed_del(self: Any) -> None:  # pragma: no cover
    """Override DataFrame's __del__ method: call the standard method + release free memory with malloc_trim."""
    if __std_del is not None:
        __std_del(self)
    if libc is not None:
        libc.malloc_trim(0)


# Override standard pandas method if needed
if libc is not None:
    pd.DataFrame.__del__ = __fixed_del



import pandas as pd

from melusine.base import MelusineDetector


# --8<-- [start:detector]
# --8<-- [start:detector_init]
class MyVirusDetector(MelusineDetector):
    """
    Detect if the text expresses dissatisfaction.
    """

    # Dataframe column names
    OUTPUT_RESULT_COLUMN = "virus_result"
    TMP_DETECTION_INPUT_COLUMN = "detection_input"
    TMP_POSITIVE_REGEX_MATCH = "positive_regex_match"
    TMP_NEGATIVE_REGEX_MATCH = "negative_regex_match"

    def __init__(self, body_column: str, header_column: str):
        self.body_column = body_column
        self.header_column = header_column

        super().__init__(
            input_columns=[self.body_column, self.header_column],
            output_columns=[self.OUTPUT_RESULT_COLUMN],
            name="virus",
        )
        # --8<-- [end:detector_init]

    # --8<-- [start:pre_detect]
    def pre_detect(self, df, debug_mode=False):
        # Assemble the text columns into a single column
        df[self.TMP_DETECTION_INPUT_COLUMN] = df[self.header_column] + "\n" + df[self.body_column]

        return df
        # --8<-- [end:pre_detect]

    # --8<-- [start:detect]
    def detect(self, df, debug_mode=False):
        text_column = df[self.TMP_DETECTION_INPUT_COLUMN]
        positive_regex = r"(virus)"
        negative_regex = r"(corona[ _]virus)"

        # Pandas str.extract method on columns
        df[self.TMP_POSITIVE_REGEX_MATCH] = text_column.str.extract(positive_regex).apply(pd.notna)
        df[self.TMP_NEGATIVE_REGEX_MATCH] = text_column.str.extract(negative_regex).apply(pd.notna)

        return df
        # --8<-- [end:detect]

    # --8<-- [start:post_detect]
    def post_detect(self, df, debug_mode=False):
        # Boolean operation on pandas column
        df[self.OUTPUT_RESULT_COLUMN] = df[self.TMP_POSITIVE_REGEX_MATCH] & ~df[self.TMP_NEGATIVE_REGEX_MATCH]
        return df

    # --8<-- [end:post_detect]


# --8<-- [end:detector]


def run():
    # --8<-- [start:run]
    df = pd.DataFrame(
        [
            {"body": "This is a dangerous virus", "header": "test"},
            {"body": "test", "header": "test"},
            {"body": "test", "header": "viruses are dangerous"},
            {"body": "corona virus is annoying", "header": "test"},
        ]
    )

    detector = MyVirusDetector(body_column="body", header_column="header")

    df = detector.transform(df)
    # --8<-- [end:run]

    return df



import pytest

from melusine.pipeline import MelusinePipeline


@pytest.fixture
def conf_pipeline_basic():
    return {
        "test_pipeline": {
            "steps": [
                {
                    "class_name": "Normalizer",
                    "module": "melusine.processor",
                    "config_key": "test_normalizer",
                },
                {
                    "class_name": "RegexTokenizer",
                    "module": "melusine.processor",
                    "config_key": "test_tokenizer",
                },
            ]
        }
    }


@pytest.fixture
def pipeline_default():
    # Load json config
    conf = MelusinePipeline.get_config_from_key(config_key="my_pipeline")

    # Prevent model loading
    # (Set all model_name parameters to None)
    for step in conf["steps"]:
        step["name"] = step.pop("config_key", None)
        params = step["parameters"]

        if "model_name" in params:
            params["model_name"] = None

    # Create pipeline from a json config file (using config key "my_pipeline")
    return MelusinePipeline.from_config(config_dict=conf, verbose=True)



"""
Backend to run transforms on dict objects.

Implemented classes: [
    DictBackend,
]
"""

from typing import Any, Callable, Dict, List, Optional

from melusine.backend.base_backend import BaseTransformerBackend


class DictBackend(BaseTransformerBackend):
    """
    Backend class to operate on dict objects.
    Inherits from the BaseTransformerBackend abstract class.
    """

    def apply_transform(
        self,
        data: Dict[str, Any],
        func: Callable,
        output_columns: Optional[List[str]] = None,
        input_columns: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """
        Method to apply a transform on a Dataset using the Dict backend.

        Parameters
        ----------
        data: Dict[str: Any]
            Data to be transformed
        func: Callable
            Transform function to apply to the input data
        output_columns: List[str]
            List of output columns
        input_columns: List[str]
            List of input columns
        kwargs

        Returns
        -------
        _: Dict[str: Any]
            Transformed data
        """
        if input_columns and len(input_columns) == 1:
            input_column = input_columns[0]

            # Modify the entire dict
            if not output_columns:
                raise ValueError("DictBackend does not support single input + None output situation.")

            # Create a single new field
            elif len(output_columns) == 1:
                output_column = output_columns[0]
                data[output_column] = func(data[input_column], **kwargs)

            # Create multiple new fields
            else:
                result = func(data[input_column], **kwargs)
                data.update(dict(zip(output_columns, result)))

        # Use DataFrame.apply
        else:
            # Modify the entire dict
            if not output_columns:
                data = func(data, **kwargs)

            # Create a single new field
            elif len(output_columns) == 1:
                output_column = output_columns[0]
                data[output_column] = func(data, **kwargs)

            # Create multiple new fields
            else:
                result = func(data, **kwargs)
                data.update(dict(zip(output_columns, result)))

        return data

    def add_fields(
        self, left: Dict[str, Any], right: Dict[str, Any], fields: Optional[List[str]] = None
    ) -> Dict[str, Any]:
        """
        Method to add fields form the right object to the left object.

        Parameters
        ----------
        left: Dict[str, Any]
            MelusineDataset object
        right: Dict[str, Any]
            Melusine Dataset object
        fields: List[str]
            List of fields to be added

        Returns
        -------
        _: Dict[str, Any]
            Left object with added fields
        """
        if not fields:
            fields = list(right.keys())

        for field in fields:
            left[field] = right[field]

        return left

    def copy(self, data: Dict[str, Any], fields: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        Method to make a copy of the dataset.

        Parameters
        ----------
        data: Dict[str, Any]
            MelusineDataset object
        fields: List[str]
            List of fields to include in the copy (by default copy all fields)

        Returns
        -------
        _: Dict[str, Any]
            Copy of original object
        """
        new_dict = dict()

        if fields is None:
            fields = list(data.keys())

        for field in fields:
            new_dict[field] = data[field]

        return new_dict

    def get_fields(self, data: Dict[str, Any]) -> List[str]:
        """
        Method to get the list of fields available in the input dataset.

        Parameters
        ----------
        data: Dict[str, Any]
            MelusineDataset object

        Returns
        -------
        _: List[str]
            List of dataset fields
        """
        return list(data.keys())

    def setup_debug_dict(self, data: Dict[str, Any], dict_name: str) -> Dict[str, Any]:
        """
        Method to check if debug_mode is activated.

        Parameters
        ----------
        data: Dict[str, Any]
            MelusineDataset object
        dict_name: str
            Name of the debug dict field to be added

        Returns
        -------
        _: Dict[str, Any]
            MelusineDataset object
        """
        data[dict_name] = {}

        return data



import sys
from pathlib import Path
from typing import Generator

import pytest

# Package source root
docs_folder = Path(__file__).parents[2] / "docs"


@pytest.fixture
def add_docs_to_pythonpath() -> Generator[None, None, None]:
    """Testing"""
    # Add docs to python path
    sys.path.insert(0, str(docs_folder))
    yield None



from typing import Dict, List, Optional, Union

from melusine.base import MelusineRegex


class EmergencyRegex(MelusineRegex):
    """
    Detect reply patterns in headers such as "re:".
    """

    @property
    def positive(self) -> Union[str, Dict[str, str]]:
        """
        Define regex patterns required to activate the MelusineRegex.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return r"urgent|emergency"

    @property
    def neutral(self) -> Optional[Union[str, Dict[str, str]]]:
        """
        Define regex patterns to be ignored when running detection.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return r"emergency exit"

    @property
    def negative(self) -> Optional[Union[str, Dict[str, str]]]:
        """
        Define regex patterns prohibited to activate the MelusineRegex.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return dict(
            blacklist=r"Mrs. TooInsistent|Mr. Annoying",
            not_my_business=r"GalaxyFarFarAway",
        )

    @property
    def match_list(self) -> List[str]:
        """
        List of texts that should activate the MelusineRegex.

        Returns:
            _: List of texts.
        """
        return [
            "We have an emergency",
            "This message is urgent",
        ]

    @property
    def no_match_list(self) -> List[str]:
        """
        List of texts that should NOT activate the MelusineRegex.

        Returns:
            _: List of texts.
        """
        return [
            "Mr. Annoying is calling for an emergency",
            "Mrs. TooInsistent called 8 times for an urgent matter",
            "There is an emergency in GalaxyFarFarAway",
        ]



"""
Data container class for email.

An email body can contain many "messages".

Implemented classes: [Message]
"""
import re
from datetime import datetime
from typing import Iterable, List, Optional, Tuple

from melusine import config


class Message:
    """
    Class acting as a data container for email data (text, meta and features)
    """

    DEFAULT_STR_LINE_LENGTH = 120
    DEFAULT_STR_TAG_NAME_LENGTH = 22

    def __init__(
        self,
        text: str,
        header: str = "",
        meta: str = "",
        date: Optional[datetime] = None,
        text_from: str = "",
        text_to: Optional[str] = None,
        tags: Optional[List[Tuple[str, str]]] = None,
    ):
        """
        Attributes initialization.

        Parameters
        ----------
        text: str
            Message text content.
        header: str
            Message text header.
        meta: str
            Message raw metadata.
        date: datetime
            Message date.
        text_from: str
            Email sender.
        text_to: str
            Email receiver.
        tags: List[Tuple[str, str]]
            Tagged test parts.
            (should be passed as init argument for debug purposes only)
        """
        self.text = text
        self.header = header
        self.meta = meta
        self.date = date
        self.text_from = text_from
        self.text_to = text_to

        self.tags = tags
        self.clean_header: str = ""
        self.clean_text: str = ""

    @property
    def str_tag_name_length(self) -> int:
        """
        When printing a message, number of characters for the TAG field.
        """
        if "message" not in config:
            return self.DEFAULT_STR_TAG_NAME_LENGTH
        else:
            return config["message"].get("str_tag_name_length", self.DEFAULT_STR_TAG_NAME_LENGTH)

    @property
    def str_line_length(self) -> int:
        """
        When printing a message, total number of characters in each line (text + separation + tag).
        """
        if "message" not in config:
            return self.DEFAULT_STR_LINE_LENGTH
        else:
            return config["message"].get("str_line_length", self.DEFAULT_STR_LINE_LENGTH)

    def extract_parts(
        self, target_tags: Optional[Iterable[str]] = None, stop_at: Optional[Iterable[str]] = None
    ) -> List[Tuple[str, str]]:
        """
        Function to extract target tags from the message.

        Parameters
        ----------
        target_tags:
            Tags to be extracted.
        stop_at:
            Tags for which extraction should stop.

        Returns
        -------
        _: List[Tuple[str, str]]
            List of extracted tags.
        """
        if not self.tags:
            return []

        # List of tags in the message
        tag_name_list: List[str] = [x[0] for x in self.tags]

        if target_tags is None:
            target_tags = tag_name_list

        # When stop tags are specified, work on a restricted message
        # (Ex: All tags until GREETINGS)
        if stop_at:
            upper_bound: int = len(tag_name_list)
            for tag_name in stop_at:
                if tag_name in tag_name_list:
                    upper_bound = min(upper_bound, tag_name_list.index(tag_name))
            # Restrict message
            effective_tags = self.tags[:upper_bound]
        else:
            effective_tags = self.tags

        return [x for x in effective_tags if x[0] in target_tags]

    def extract_last_body(
        self, target_tags: Iterable[str] = ("BODY",), stop_at: Iterable[str] = ("GREETINGS",)
    ) -> List[Tuple[str, str]]:
        """
        Extract the BODY parts of the last message in the email.

        Parameters
        ----------
        target_tags: Iterable[str]
        stop_at: Iterable[str]

        Returns
        -------
        _: List[Tuple[str, str]]
        """
        return self.extract_parts(target_tags=target_tags, stop_at=stop_at)

    def has_tags(
        self,
        target_tags: Iterable[str] = ("BODY",),
        stop_at: Optional[Iterable[str]] = None,
    ) -> bool:
        """
        Function to check if input tags are present in the message.

        Parameters
        ----------
        target_tags:
            Tags of interest.
        stop_at:
            Tags for which extraction should stop.

        Returns
        -------
        _: bool
            True if target tags are present in the message.
        """
        if self.tags is None:
            return False

        if not stop_at:
            stop_at = set()

        found: bool = False
        for tag, _ in self.tags:
            # Check if tag in tags of interest
            if tag in target_tags:
                found = True
                break

            # Stop when specified tag is reached
            if tag in stop_at:
                break

        return found

    def format_tags(self) -> str:
        """
        Create a pretty formatted representation of text and their associated tags.

        Returns:
            _: Pretty formatted representation of the tags and texts.
        """
        if self.tags is None:
            return self.text
        else:
            tag_text_length = self.str_line_length - self.str_tag_name_length
            text = ""
            for tag_name, tag_text in self.tags:
                text += tag_text.ljust(tag_text_length, ".") + tag_name.rjust(self.str_tag_name_length, ".") + "\n"

        return text.strip()

    def __repr__(self) -> str:
        """
        String representation.

        Returns
        -------
        _: str
            Readable representation of the Message.
        """
        if self.meta:
            meta = re.sub(r"\n+", r"\n", self.meta).strip("\n ")
        else:
            meta = "NA"
        text: str = re.sub(r"\n+", r"\n", self.text)
        return f"Message(meta={repr(meta)}, text={repr(text)})"

    def __str__(self) -> str:
        """
        Repr representation.

        Returns
        -------
        _: str
            Readable representation of the Message.
        """
        title_len = 22
        fill_len = (self.str_line_length - title_len) // 2

        text = ""
        text += f"{'='*fill_len}{'Message':^{title_len}}{'='*fill_len}\n"
        text += f"{'-'*fill_len}{'Meta':^{title_len}}{'-'*fill_len}\n"
        text += f"{self.meta or 'N/A'}\n"
        text += f"{'-'*fill_len}{'Text':^{title_len}}{'-'*fill_len}\n"
        text += self.format_tags() + "\n"
        text += f"{'='*fill_len}{'=' * title_len}{'='*fill_len}\n\n"

        return text



def test_tutorial001(add_docs_to_pythonpath):
    from docs_src.GettingStarted.tutorial001 import run as run001

    _ = run001()


def test_tutorial002(add_docs_to_pythonpath):
    from docs_src.GettingStarted.tutorial001 import run as run002

    _ = run002()



"""
The melusine.regex module includes tools for handling regexes.
"""
from melusine.regex.emergency_regex import EmergencyRegex
from melusine.regex.reply_regex import ReplyRegex
from melusine.regex.thanks_regex import ThanksRegex
from melusine.regex.transfer_regex import TransferRegex
from melusine.regex.vacation_reply_regex import VacationReplyRegex

__all__ = ["EmergencyRegex", "ReplyRegex", "ThanksRegex", "TransferRegex", "VacationReplyRegex"]



import logging
import os
import re

import pytest

from melusine import config
from melusine._config import MelusineConfig, MelusineConfigError


def test_load_default_conf(caplog):
    with caplog.at_level(logging.WARNING):
        test_conf = MelusineConfig()
        test_conf.reset()

        assert test_conf
        assert MelusineConfig.LOG_MESSAGE_DEFAULT_CONFIG in caplog.text


def test_load_conf_from_env_variable(caplog):
    try:
        with caplog.at_level(logging.INFO):
            test_conf = MelusineConfig()
            os.environ[MelusineConfig.ENV_MELUSINE_CONFIG_DIR] = test_conf.DEFAULT_CONFIG_PATH
            test_conf.reset()

            expected_config_path_log = MelusineConfig.LOG_MESSAGE_CONFIG_PATH.format(
                config_path=test_conf.DEFAULT_CONFIG_PATH
            )

            assert test_conf
            assert MelusineConfig.LOG_MESSAGE_CONFIG_FROM_ENV_VARIABLE in caplog.text
            assert expected_config_path_log in caplog.text
            assert MelusineConfig.LOG_MESSAGE_DEFAULT_CONFIG not in caplog.text

    finally:
        del os.environ[MelusineConfig.ENV_MELUSINE_CONFIG_DIR]


def test_load_conf_from_config_path(caplog):
    with caplog.at_level(logging.INFO):
        test_conf = MelusineConfig()
        test_conf.reset(config_path=test_conf.DEFAULT_CONFIG_PATH)

        expected_config_path_log = MelusineConfig.LOG_MESSAGE_CONFIG_PATH.format(
            config_path=test_conf.DEFAULT_CONFIG_PATH
        )

        assert test_conf
        assert expected_config_path_log in caplog.text
        assert MelusineConfig.LOG_MESSAGE_CONFIG_FROM_ENV_VARIABLE not in caplog.text
        assert MelusineConfig.LOG_MESSAGE_DEFAULT_CONFIG not in caplog.text


def test_load_conf_from_config_dict(caplog):
    with caplog.at_level(logging.INFO):
        test_conf = MelusineConfig()
        test_conf.reset(config_dict={"my_key": "hello"})

        assert test_conf["my_key"] == "hello"


def test_config_modif_error():
    test_conf = MelusineConfig()
    test_conf.reset(config_dict={"my_key": "hello"})

    with pytest.raises(MelusineConfigError, match=re.escape(MelusineConfigError.CONST_CONFIG_ERROR_MESSAGE)):
        test_conf["new_key"] = "hey"

    with pytest.raises(MelusineConfigError):
        test_conf.pop()

    with pytest.raises(MelusineConfigError):
        test_conf.popitem()


def test_shared_variable():
    # Shared variable TEST_VAR specified in conf/shared.yaml
    # Conf test_shared_variable specified in global.yaml
    assert config["global"]["test_shared_variable"] == "test"


def test_export_config(tmp_path):
    file_list = config.export_default_config(path=str(tmp_path))
    assert file_list
    for file in file_list:
        assert file.endswith(".yaml")



from typing import Dict, List, Optional, Union

from melusine.base import MelusineRegex


class ThanksRegex(MelusineRegex):
    """
    Detect thanks patterns such as "merci".
    """

    @property
    def positive(self) -> Union[str, Dict[str, str]]:
        """
        Define regex patterns required to activate the MelusineRegex.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return r"\bmerci+s?\b|\bremercie?"

    @property
    def neutral(self) -> Optional[Union[str, Dict[str, str]]]:
        """
        Define regex patterns to be ignored when running detection.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return None

    @property
    def negative(self) -> Optional[Union[str, Dict[str, str]]]:
        """
        Define regex patterns prohibited to activate the MelusineRegex.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        forbidden_thanks_words = [
            r"oui",
            r"non",
            r"atten[dt]",
            r"inform[ée]",
            r"proposition",
            r"ci\b",
            r"join[st]?\b",
        ]

        return dict(QUESTION=r"\?", FORBIDDEN_WORDS=r"\b(" + "|".join(forbidden_thanks_words) + ")")

    @property
    def match_list(self) -> List[str]:
        """
        List of texts that should activate the MelusineRegex.

        Returns:
            _: List of texts.
        """
        return [
            "merci",
            "Je vous remercie pour votre rapidité",
            "un grand MERCI à la MAIF",
            "je tiens à remercier l'équipe",
        ]

    @property
    def no_match_list(self) -> List[str]:
        """
        List of texts that should NOT activate the MelusineRegex.

        Returns:
            _: List of texts.
        """
        return [
            # réponse à une question ouverte
            "oui, merci à vous",
            "non, merci quand même",
            "merci, j'attends votre réponse",
            "j'aimerais être tenu informée merci",
            "Merci, faites moi une proposition",
            "Merci, ci-joint le formulaire",
            "ci-attaché ledocument, merci",
            "Madame Mercier",
        ]



"""
Module which handles the package configuration.
"""
import copy
import logging
import os
from collections import UserDict
from pathlib import Path
from typing import Any, Dict, List, Optional, cast, no_type_check

from omegaconf import OmegaConf

logger = logging.getLogger(__name__)
CONST_ENV_MELUSINE_CONFIG_DIR = "MELUSINE_CONFIG_DIR"


class MelusineConfig(UserDict):
    """
    The MelusineConfig class acts as a dict containing configurations.
    The configurations can be changed dynamically using the switch_config function.
    """

    ENV_MELUSINE_CONFIG_DIR = "MELUSINE_CONFIG_DIR"
    LOG_MESSAGE_DEFAULT_CONFIG = "Using default configurations."
    LOG_MESSAGE_CONFIG_FROM_ENV_VARIABLE = f"Using config_path from env variable {ENV_MELUSINE_CONFIG_DIR}."
    LOG_MESSAGE_CONFIG_PATH = "Using config_path : {config_path}."
    DEFAULT_CONFIG_PATH = str(Path(__file__).parent.resolve() / "conf")

    @no_type_check
    def pop(self, s: Any = None) -> None:
        """
        Prevent MelusineConfig modification.
        """
        raise MelusineConfigError()

    @no_type_check
    def popitem(self, s: Any = None) -> None:
        """
        Prevent MelusineConfig modification.
        """
        raise MelusineConfigError()

    def __setitem__(self, key: str, value: Any) -> None:
        """
        Prevent MelusineConfig modification.
        """
        raise MelusineConfigError()

    def dict(self) -> Dict[str, Any]:
        """
        Return a copy of the config dict.
        """
        return copy.deepcopy(self.data)

    @staticmethod
    def _load_from_path(config_path: str) -> Dict[str, Any]:
        """
        Load yaml config files, merge them and return a config dict.
        """
        yaml_conf_file_list = list(Path(config_path).rglob("*.yaml")) + list(Path(config_path).rglob("*.yml"))
        omega_conf = OmegaConf.unsafe_merge(*[OmegaConf.load(conf_file) for conf_file in yaml_conf_file_list])
        return cast(Dict[str, Any], OmegaConf.to_object(omega_conf))

    def reset(self, config_dict: Optional[Dict[str, Any]] = None, config_path: Optional[str] = None) -> None:
        """
        Function to reset the Melusine configuration using a dict or a path.

        Parameters
        ----------
            config_dict: Dict containing the new config.
            config_path: path to directory containing YAML config files.
        """
        config_path_from_env = os.getenv(self.ENV_MELUSINE_CONFIG_DIR)

        if not config_dict and not config_path:
            if config_path_from_env:
                logger.info(self.LOG_MESSAGE_CONFIG_FROM_ENV_VARIABLE)
                config_path = config_path_from_env
            else:
                logger.warning(self.LOG_MESSAGE_DEFAULT_CONFIG)
                config_dict = self._load_from_path(self.DEFAULT_CONFIG_PATH)

        if config_path:
            logger.info(self.LOG_MESSAGE_CONFIG_PATH.format(config_path=config_path))
            config_dict = self._load_from_path(config_path)

        if config_dict is None:
            raise MelusineConfigError()  # pragma no cover

        self.data = config_dict

    def export_default_config(self, path: str) -> List[str]:
        """
        Export the default Melusine configurations to a directory.

        Parameters
        ----------
        path: Destination path

        Returns
        -------
        _: the list of all copied files.
        """
        from shutil import copytree

        source = self.DEFAULT_CONFIG_PATH
        new_directory: str = copytree(source, path, dirs_exist_ok=True)
        copied_files = [str(path) for path in Path(new_directory).rglob("*") if path.is_file()]

        return copied_files


# Load Melusine configuration
config = MelusineConfig()
config.reset()


class MelusineConfigError(Exception):
    """
    Exception raised when encountering config related errors.
    """

    CONST_CONFIG_ERROR_MESSAGE = f"""To modify the config use the `reset` method:
    - Using a dict:
      > from melusine import config
      > config.reset(config_dict=my_dict)
    - Using the path to a directory containing YAML files:
      > from melusine import config
      > config.reset(config_path=my_config_path)
    - Reset to default configurations:
      > from melusine import config
      > config.reset()
    - Using the {MelusineConfig.ENV_MELUSINE_CONFIG_DIR} environment variable:
      > import os
      > os.environ["{MelusineConfig.ENV_MELUSINE_CONFIG_DIR}"] = "/path/to/config/dir"
      > from melusine import config
      > config.reset()
    """

    def __init__(self, msg: str = CONST_CONFIG_ERROR_MESSAGE, *args: Any) -> None:
        """
        Initialize with a default error message.
        """
        super().__init__(msg, *args)
